import sys
from os.path import abspath, dirname, join

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import (
    Button,
    ButtonWidget,
    Calendar,
    CommunityList,
    CustomWidget,
    IDCard,
    Image,
    ImageData,
    ImageWidget,
    Menu,
    MenuLink,
    ModeratorsWidget,
    PostFlairWidget,
    Redditor,
    RulesWidget,
    Submenu,
    Subreddit,
    TextArea,
    Widget,
)

from ... import IntegrationTest


def image_path(name):
    test_dir = abspath(dirname(sys.modules[__name__].__file__))
    return abspath(join(test_dir, "..", "..", "files", name))


class TestButtonWidget(IntegrationTest):
    async def test_button_widget(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            button_widget = None
            async for widget in widgets.sidebar():
                if isinstance(widget, ButtonWidget):
                    button_widget = widget
                    break
            assert isinstance(button_widget, ButtonWidget)
            assert len(button_widget) >= 1
            assert all(isinstance(button, Button) for button in button_widget.buttons)
            assert button_widget == button_widget
            assert button_widget.id == button_widget
            assert button_widget in await self.async_list(widgets.sidebar())

            assert button_widget[0].text
            assert button_widget.shortName
            assert hasattr(button_widget, "description")

            assert subreddit == button_widget.subreddit

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            styles = {"headerColor": "#123456", "backgroundColor": "#bb0e00"}
            my_image = await widgets.mod.upload_image(image_path("test.png"))
            buttons = [
                {
                    "kind": "text",
                    "text": "View source",
                    "url": "https://github.com/praw-dev/asyncpraw",
                    "color": "#FF0000",
                    "textColor": "#00FF00",
                    "fillColor": "#0000FF",
                    "hoverState": {
                        "kind": "text",
                        "text": "VIEW SOURCE!",
                        "color": "#FFFFFF",
                        "textColor": "#000000",
                        "fillColor": "#0000FF",
                    },
                },
                {
                    "kind": "image",
                    "text": "View documentation",
                    "linkUrl": "https://asyncpraw.readthedocs.io",
                    "url": my_image,
                    "height": 200,
                    "width": 200,
                    "hoverState": {
                        "kind": "image",
                        "url": my_image,
                        "height": 200,
                        "width": 200,
                    },
                },
                {
                    "kind": "text",
                    "text": "/r/redditdev",
                    "url": "https://reddit.com/r/redditdev",
                    "color": "#000000",
                    "textColor": "#FF00FF",
                    "fillColor": "#005500",
                },
            ]
            widget = await widgets.mod.add_button_widget(
                short_name="Things to click",
                description="Click some of these *cool* links!",
                buttons=buttons,
                styles=styles,
            )

            assert isinstance(widget, ButtonWidget)
            assert len(widget) == 3
            assert all(isinstance(item, Button) for item in widget)
            assert widget.shortName == "Things to click"
            assert widget.description == "Click some of these *cool* links!"
            assert widget.styles == styles

            assert widget[0].text == "View source"
            assert widget[0].url == "https://github.com/praw-dev/asyncpraw"
            assert widget[2].text == "/r/redditdev"
            assert widget[2].url == "https://reddit.com/r/redditdev"

            assert widget[1].text == "View documentation"
            assert widget[1].linkUrl == "https://asyncpraw.readthedocs.io"
            assert widget[1].hoverState["kind"] == "image"
            assert widget[1].hoverState["height"] == 200

            await widgets.refresh()  # the links are initially invalid
            async for new_widget in widgets.sidebar():
                if new_widget == widget:
                    widget = new_widget
                    break

            widget = await widget.mod.update(shortName="New short name")

            assert isinstance(widget, ButtonWidget)
            assert len(widget) == 3
            assert all(isinstance(item, Button) for item in widget)
            assert widget.shortName == "New short name"
            assert widget.description == "Click some of these *cool* links!"
            assert widget.styles == styles

            assert widget[0].text == "View source"
            assert widget[0].url == "https://github.com/praw-dev/asyncpraw"
            assert widget[2].text == "/r/redditdev"
            assert widget[2].url == "https://reddit.com/r/redditdev"

            assert widget[1].text == "View documentation"
            assert widget[1].linkUrl == "https://asyncpraw.readthedocs.io"
            assert widget[1].hoverState["kind"] == "image"
            assert widget[1].hoverState["height"] == 200

            buttons.reverse()
            widget = await widget.mod.update(buttons=buttons)

            assert isinstance(widget, ButtonWidget)
            assert len(widget) == 3
            assert all(isinstance(item, Button) for item in widget)
            assert widget.shortName == "New short name"
            assert widget.description == "Click some of these *cool* links!"
            assert widget.styles == styles

            assert widget[0].text == "/r/redditdev"
            assert widget[0].url == "https://reddit.com/r/redditdev"
            assert widget[2].text == "View source"
            assert widget[2].url == "https://github.com/praw-dev/asyncpraw"

            assert widget[1].text == "View documentation"
            assert widget[1].linkUrl == "https://asyncpraw.readthedocs.io"
            assert widget[1].hoverState["kind"] == "image"
            assert widget[1].hoverState["height"] == 200

            await widget.mod.delete()


class TestCalendar(IntegrationTest):
    async def test_calendar(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            calendar = None
            async for widget in widgets.sidebar():
                if isinstance(widget, Calendar):
                    calendar = widget
                    break
            assert isinstance(calendar, Calendar)
            assert calendar == calendar
            assert calendar.id == calendar
            assert calendar in await self.async_list(widgets.sidebar())

            assert isinstance(calendar.configuration, dict)
            assert hasattr(calendar, "requiresSync")

            assert subreddit == calendar.subreddit

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            styles = {"headerColor": "#123456", "backgroundColor": "#bb0e00"}
            config = {
                "numEvents": 10,
                "showDate": True,
                "showDescription": False,
                "showLocation": False,
                "showTime": True,
                "showTitle": True,
            }
            cal_id = "ccahu0rstno2jrvioq4ccffn78@group.calendar.google.com"
            widget = await widgets.mod.add_calendar(
                short_name="Upcoming Events",
                google_calendar_id=cal_id,
                requires_sync=True,
                configuration=config,
                styles=styles,
            )

            assert isinstance(widget, Calendar)
            assert widget.shortName == "Upcoming Events"
            assert (
                widget.googleCalendarId
                == "ccahu0rstno2jrvioq4ccffn78@group.calendar.google.com"
            )
            assert widget.configuration == config
            assert widget.styles == styles

            widget = await widget.mod.update(shortName="Past Events :(")

            assert isinstance(widget, Calendar)
            assert widget.shortName == "Past Events :("
            assert (
                widget.googleCalendarId
                == "ccahu0rstno2jrvioq4ccffn78@group.calendar.google.com"
            )
            assert widget.configuration == config
            assert widget.styles == styles

            await widget.mod.delete()


class TestCommunityList(IntegrationTest):
    async def test_community_list(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            comm_list = None
            async for widget in widgets.sidebar():
                if isinstance(widget, CommunityList):
                    comm_list = widget
                    break
            assert isinstance(comm_list, CommunityList)
            assert len(comm_list) >= 1
            assert all(isinstance(subreddit, Subreddit) for subreddit in comm_list)
            assert comm_list == comm_list
            assert comm_list.id == comm_list
            assert comm_list in await self.async_list(widgets.sidebar())

            assert comm_list.shortName
            assert comm_list[0] in comm_list

            assert subreddit == comm_list.subreddit

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            styles = {"headerColor": "#123456", "backgroundColor": "#bb0e00"}
            subreddits = ["learnpython", Subreddit(self.reddit, "redditdev")]
            widget = await widgets.mod.add_community_list(
                short_name="My fav subs",
                data=subreddits,
                styles=styles,
                description="My description",
            )
            assert isinstance(widget, CommunityList)
            assert widget.shortName == "My fav subs"
            assert widget.styles == styles
            # assert widget.description == "My description" # no longer returned
            assert Subreddit(self.reddit, "learnpython") in widget
            assert "redditdev" in widget

            widget = await widget.mod.update(
                shortName="My least fav subs :(",
                data=["redesign"],
                description="My new description",
            )

            assert isinstance(widget, CommunityList)
            assert widget.shortName == "My least fav subs :("
            assert widget.styles == styles
            # assert widget.description == "My new description" # no longer returned
            assert Subreddit(self.reddit, "redesign") in widget

            await widget.mod.delete()


class TestCustomWidget(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            image_dicts = [
                {
                    "width": 0,
                    "height": 0,
                    "name": "a",
                    "url": await widgets.mod.upload_image(image_path("test.png")),
                }
            ]

            styles = {"headerColor": "#123456", "backgroundColor": "#bb0e00"}
            widget = await widgets.mod.add_custom_widget(
                short_name="My widget",
                text="# Hello world!",
                css="/**/",
                height=200,
                image_data=image_dicts,
                styles=styles,
            )

            assert isinstance(widget, CustomWidget)
            assert widget.shortName == "My widget"
            assert widget.text == "# Hello world!"
            assert widget.css == "/**/"
            assert widget.height == 200
            assert widget.styles == styles
            assert len(widget.imageData) == 1
            assert all(isinstance(img, ImageData) for img in widget.imageData)

            # initially, image URLs are incorrect, so we much refresh to get
            # the proper ones.
            await widgets.refresh()
            sidebar = await self.async_list(widgets.sidebar())
            refreshed = sidebar[-1]
            assert refreshed == widget
            widget = refreshed

            new_css = "h1,h2,h3,h4,h5,h6 {color: #00ff00;}"
            widget = await widget.mod.update(css=new_css)

            assert isinstance(widget, CustomWidget)
            assert widget.shortName == "My widget"
            assert widget.text == "# Hello world!"
            assert widget.css == new_css
            assert widget.height == 200
            assert widget.styles == styles
            assert len(widget.imageData) == 1
            assert all(isinstance(img, ImageData) for img in widget.imageData)

            await widget.mod.delete()

    async def test_custom_widget(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            custom = None
            async for widget in widgets.sidebar():
                if isinstance(widget, CustomWidget):
                    custom = widget
                    break
            assert isinstance(custom, CustomWidget)
            assert len(custom.imageData) > 0
            assert all(isinstance(img_data, ImageData) for img_data in custom.imageData)
            assert custom == custom
            assert custom.id == custom
            assert custom in await self.async_list(widgets.sidebar())

            assert 500 >= custom.height >= 50
            assert custom.text
            assert custom.css
            assert custom.shortName

            assert subreddit == custom.subreddit


class TestIDCard(IntegrationTest):
    async def test_id_card(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            card = await widgets.id_card()
            assert isinstance(card, IDCard)
            assert card == card
            assert card.id == card

            assert card.shortName
            assert card.currentlyViewingText
            assert card.subscribersText

            assert subreddit == card.subreddit

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette():
            id_card = await widgets.id_card()
            assert id_card.currentlyViewingText != "Users here NOW!"
            assert id_card.subscribersText != "Loyal readers"

            await id_card.mod.update(currentlyViewingText="Users here NOW!")
            await widgets.refresh()
            id_card = await widgets.id_card()

            assert id_card.currentlyViewingText == "Users here NOW!"
            assert id_card.subscribersText != "Loyal readers"

            await id_card.mod.update(subscribersText="Loyal readers")
            await widgets.refresh()
            id_card = await widgets.id_card()

            assert id_card.currentlyViewingText == "Users here NOW!"
            assert id_card.subscribersText == "Loyal readers"


class TestImageWidget(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            image_paths = (image_path(name) for name in ("test.jpg", "test.png"))
            image_data = [
                {
                    "width": 10,
                    "height": 10,
                    "linkUrl": "",
                    "url": await widgets.mod.upload_image(img_path),
                }
                for img_path in image_paths
            ]

            styles = {"headerColor": "#123456", "backgroundColor": "#bb0e00"}
            widget = await widgets.mod.add_image_widget(
                short_name="My new pics!", data=image_data, styles=styles
            )

            assert isinstance(widget, ImageWidget)
            assert widget.shortName == "My new pics!"
            assert widget.styles == styles
            assert len(widget) == 2
            assert all(isinstance(img, Image) for img in widget)

            widget = await widget.mod.update(
                shortName="My old pics :(", data=image_data[:1]
            )

            assert isinstance(widget, ImageWidget)
            assert widget.shortName == "My old pics :("
            assert widget.styles == styles
            assert len(widget) == 1
            assert all(isinstance(img, Image) for img in widget)

            await widget.mod.delete()

    async def test_image_widget(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            img_widget = None
            async for widget in widgets.sidebar():
                if isinstance(widget, ImageWidget):
                    img_widget = widget
                    break
            assert isinstance(img_widget, ImageWidget)
            assert len(img_widget) >= 1
            assert all(isinstance(image, Image) for image in img_widget)
            assert img_widget == img_widget
            assert img_widget.id == img_widget
            assert img_widget in await self.async_list(widgets.sidebar())

            assert img_widget[0].linkUrl
            assert img_widget.shortName

            assert subreddit == img_widget.subreddit


class TestMenu(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        menu_contents = [
            {"text": "My homepage", "url": "https://example.com"},
            {
                "text": "Python packages",
                "children": [
                    {"text": "PRAW", "url": "https://asyncpraw.readthedocs.io/"},
                    {"text": "requests", "url": "http://python-requests.org"},
                ],
            },
            {"text": "Reddit homepage", "url": "https://reddit.com"},
        ]

        with self.use_cassette():
            widget = await widgets.mod.add_menu(data=menu_contents)

            assert isinstance(widget, Menu)
            assert len(widget) == 3
            assert all(isinstance(item, (Submenu, MenuLink)) for item in widget)
            assert all(
                all(isinstance(item, MenuLink) for item in subm)
                for subm in widget
                if isinstance(subm, Submenu)
            )

            assert widget[0].text == "My homepage"
            assert widget[0].url == "https://example.com"
            assert widget[2].text == "Reddit homepage"
            assert widget[2].url == "https://reddit.com"

            assert widget[1].text == "Python packages"
            assert widget[1][0].text == "PRAW"
            assert widget[1][0].url == "https://asyncpraw.readthedocs.io/"
            assert widget[1][1].text == "requests"
            assert widget[1][1].url == "http://python-requests.org"

            menu_contents.reverse()
            widget = await widget.mod.update(data=menu_contents)

            assert isinstance(widget, Menu)
            assert len(widget) == 3
            assert all(isinstance(item, (Submenu, MenuLink)) for item in widget)
            assert all(
                all(isinstance(item, MenuLink) for item in subm)
                for subm in widget
                if isinstance(subm, Submenu)
            )

            assert widget[0].text == "Reddit homepage"
            assert widget[0].url == "https://reddit.com"
            assert widget[2].text == "My homepage"
            assert widget[2].url == "https://example.com"

            assert widget[1].text == "Python packages"
            assert widget[1][0].text == "PRAW"
            assert widget[1][0].url == "https://asyncpraw.readthedocs.io/"
            assert widget[1][1].text == "requests"
            assert widget[1][1].url == "http://python-requests.org"

            await widget.mod.delete()

    async def test_menu(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            menu = None
            async for widget in widgets.topbar():
                if isinstance(widget, Menu):
                    menu = widget
                    break
            assert isinstance(menu, Menu)
            assert all(isinstance(item, (MenuLink, Submenu)) for item in menu)
            assert menu == menu
            assert menu.id == menu
            assert menu in await self.async_list(widgets.topbar())
            assert len(menu) >= 1
            assert menu[0].text

            assert subreddit == menu.subreddit

            submenu = None
            for child in menu:
                if isinstance(child, Submenu):
                    submenu = child
                    break
            assert isinstance(submenu, Submenu)
            assert len(submenu) >= 0
            assert all(isinstance(child, MenuLink) for child in submenu)
            assert submenu[0].text
            assert submenu[0].url


class TestModeratorsWidget(IntegrationTest):
    async def test_moderators_widget(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            mods = await widgets.moderators_widget()
            assert isinstance(mods, ModeratorsWidget)
            assert all(isinstance(mod, Redditor) for mod in mods)
            assert mods == mods
            assert mods.id == mods

            assert len(mods) >= 1
            assert isinstance(mods[0], Redditor)

            assert subreddit == mods.subreddit

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        new_styles = {"backgroundColor": "#bababa", "headerColor": "#407bee"}
        with self.use_cassette():
            mod_widget = await widgets.moderators_widget()
            assert mod_widget.styles != new_styles
            await mod_widget.mod.update(styles=new_styles)
            await widgets.refresh()
            mod_widget = await widgets.moderators_widget()

            assert mod_widget.styles == new_styles


class TestPostFlairWidget(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            flairs = [f["id"] async for f in subreddit.flair.link_templates][:30]

            styles = {"headerColor": "#123456", "backgroundColor": "#bb0e00"}
            widget = await widgets.mod.add_post_flair_widget(
                display="Some flairs", order="list", short_name=flairs, styles=styles
            )

            assert isinstance(widget, PostFlairWidget)
            assert widget.shortName == "Some flairs"
            assert widget.display == "list"
            assert widget.order == flairs
            assert widget.styles == styles
            assert len(widget) == len(flairs)
            assert all(flair_id in widget.templates for flair_id in widget)

            widget = await widget.mod.update(display="cloud")

            assert isinstance(widget, PostFlairWidget)
            assert widget.shortName == "Some flairs"
            assert widget.display == "cloud"
            assert widget.order == flairs
            assert widget.styles == styles
            assert len(widget) == len(flairs)
            assert all(flair_id in widget.templates for flair_id in widget)

            widget = await widget.mod.update(order=widget.order[:1])

            assert isinstance(widget, PostFlairWidget)
            assert widget.shortName == "Some flairs"
            assert widget.display == "cloud"
            assert widget.order == flairs[:1]
            assert widget.styles == styles
            assert len(widget) == 1
            assert all(flair_id in widget.templates for flair_id in widget)

            await widget.mod.delete()

    async def test_post_flair_widget(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            pf_widget = None
            async for widget in widgets.sidebar():
                if isinstance(widget, PostFlairWidget):
                    pf_widget = widget
                    break
            assert isinstance(pf_widget, PostFlairWidget)
            assert len(pf_widget) >= 1
            assert all(flair_id in widget.templates for flair_id in widget)
            assert pf_widget == pf_widget
            assert pf_widget.id == pf_widget
            assert pf_widget in await self.async_list(widgets.sidebar())

            assert pf_widget.shortName
            assert all(flair in pf_widget for flair in pf_widget)

            assert subreddit == pf_widget.subreddit


class TestRulesWidget(IntegrationTest):
    async def test_rules_widget(self):

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            rules = None
            async for widget in widgets.sidebar():
                if isinstance(widget, RulesWidget):
                    rules = widget
                    break
            assert isinstance(rules, RulesWidget)
            assert rules == rules
            assert rules.id == rules

            assert rules.display

            assert len(rules) > 0
            assert subreddit == rules.subreddit

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        new_styles = {"backgroundColor": "#fedcba", "headerColor": "#012345"}
        with self.use_cassette():
            rules = None
            async for widget in widgets.sidebar():
                if isinstance(widget, RulesWidget):
                    rules = widget
                    break
            assert isinstance(rules, RulesWidget)

            assert rules.shortName != "Our regulations"
            assert rules.styles != new_styles

            rules = await rules.mod.update(
                display="compact",
                shortName="Our regulations",
                styles=new_styles,
            )

            assert rules.display == "compact"
            assert rules.shortName == "Our regulations"
            assert rules.styles == new_styles


class TestSubredditWidgets(IntegrationTest):
    async def test_bad_attribute(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            with pytest.raises(AttributeError):
                widgets.nonexistant_attribute

    async def test_items(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            assert isinstance(await widgets.items(), dict)

    # async def test_progressive_images(self): # FIXME: not currently working; same with praw
    #     subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
    #     widgets = subreddit.widgets
    #
    #     async def has_progressive(widgets_):
    #         # best way I could figure if an image is progressive
    #         sign = "fm=pjpg"
    #
    #         async for widget in widgets_.sidebar():
    #             if isinstance(widget, ImageWidget):
    #                 for image in widget:
    #                     if sign in image.url:
    #                         return True
    #             elif isinstance(widget, CustomWidget):
    #                 for image_data in widget.imageData:
    #                     if sign in image_data.url:
    #                         return True
    #
    #         return False
    #
    #     with self.use_cassette():
    #         widgets.progressive_images = True
    #         assert await has_progressive(widgets)
    #         widgets.progressive_images = False
    #         await widgets.refresh()
    #         assert not await has_progressive(widgets)
    #         widgets.progressive_images = True
    #         await widgets.refresh()
    #         assert await has_progressive(widgets)

    async def test_refresh(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette():
            assert await self.async_list(widgets.sidebar())  # to fetch
            old_sidebar = await self.async_list(
                widgets.sidebar()
            )  # reference, not value
            await widgets.refresh()
            new_sidebar = await self.async_list(widgets.sidebar())
            assert old_sidebar is not new_sidebar  # should be new list

    async def test_repr(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        assert (
            f"SubredditWidgets(subreddit=Subreddit(display_name='{pytest.placeholders.test_subreddit}'))"
            == repr(widgets)
        )

    async def test_sidebar(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            sidebar = [
                isinstance(widget, Widget) and type(widget) != Widget
                async for widget in widgets.sidebar()
            ]
            assert all(sidebar)

    async def test_specials(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette():
            assert isinstance(await widgets.id_card(), IDCard)
            assert isinstance(await widgets.moderators_widget(), ModeratorsWidget)

    async def test_topbar(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette():
            assert 1 <= len(await self.async_list(widgets.topbar()))
            assert all(
                [
                    isinstance(widget, Widget) and type(widget) != Widget
                    async for widget in widgets.topbar()
                ]
            )


class TestSubredditWidgetsModeration(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_reorder(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            old_order = await self.async_list(widgets.sidebar())
            new_order = list(reversed(old_order))

            await widgets.mod.reorder(new_order)
            await widgets.refresh()
            order = await self.async_list(widgets.sidebar())
            assert order == new_order

            old_order = await self.async_list(widgets.sidebar())
            new_order = list(reversed(old_order))

            await widgets.mod.reorder(new_order)
            await widgets.refresh()
            order = await self.async_list(widgets.sidebar())
            assert order == new_order

            mixed_types = [
                thing if i % 2 == 0 else thing.id for i, thing in enumerate(new_order)
            ]
            # mixed_types has some str and some Widget.
            assert any(isinstance(thing, str) for thing in mixed_types)
            assert any(isinstance(thing, Widget) for thing in mixed_types)

            await widgets.mod.reorder(mixed_types)
            await widgets.refresh()
            order = await self.async_list(widgets.sidebar())
            assert order == new_order

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_upload_image(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            for image in ("test.jpg", "test.png"):
                image_url = await widgets.mod.upload_image(image_path(image))
                assert image_url


class TestTextArea(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create_and_update_and_delete(self, _):
        self.reddit.read_only = False

        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets

        with self.use_cassette():
            styles = {"headerColor": "#123456", "backgroundColor": "#bb0e00"}
            widget = await widgets.mod.add_text_area(
                short_name="My new widget!", text="Hello world!", styles=styles
            )

            assert isinstance(widget, TextArea)
            assert widget.shortName == "My new widget!"
            assert widget.styles == styles
            assert widget.text == "Hello world!"

            widget = await widget.mod.update(
                shortName="My old widget :(", text="Feed me"
            )

            assert isinstance(widget, TextArea)
            assert widget.shortName == "My old widget :("
            assert widget.styles == styles
            assert widget.text == "Feed me"

            await widget.mod.delete()

    async def test_text_area(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            text = None
            async for widget in widgets.sidebar():
                if isinstance(widget, TextArea):
                    text = widget
                    break
            assert isinstance(text, TextArea)
            assert text == text
            assert text.id == text
            side_bar = await self.async_list(widgets.sidebar())
            assert text in side_bar

            assert text.shortName
            assert text.text

            assert subreddit == text.subreddit


class TestWidget(IntegrationTest):
    async def test_inequality(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        widgets = subreddit.widgets
        with self.use_cassette("TestSubredditWidgets.fetch_widgets"):
            sidebar = await self.async_list(widgets.sidebar())
            assert len(sidebar) >= 2
        assert sidebar[0] != sidebar[1]
        assert sidebar[0] != sidebar[1].id



"""Provide the Multireddit class."""
import re
from json import dumps
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from ...const import API_PATH
from ...util import _deprecate_args
from ...util.cache import cachedproperty
from ..listing.mixins import SubredditListingMixin
from .base import RedditBase
from .redditor import Redditor
from .subreddit import Subreddit, SubredditStream

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class Multireddit(SubredditListingMixin, RedditBase):
    r"""A class for users' multireddits.

    This is referred to as a "Custom Feed" on the Reddit UI.

    .. include:: ../../typical_attributes.rst

    ==================== ==============================================================
    Attribute            Description
    ==================== ==============================================================
    ``can_edit``         A ``bool`` representing whether or not the authenticated user
                         may edit the multireddit.
    ``copied_from``      The multireddit that the multireddit was copied from, if it
                         exists, otherwise ``None``.
    ``created_utc``      When the multireddit was created, in `Unix Time`_.
    ``description_html`` The description of the multireddit, as HTML.
    ``description_md``   The description of the multireddit, as Markdown.
    ``display_name``     The display name of the multireddit.
    ``name``             The name of the multireddit.
    ``over_18``          A ``bool`` representing whether or not the multireddit is
                         restricted for users over 18.
    ``subreddits``       A list of :class:`.Subreddit`\ s that make up the multireddit.
    ``visibility``       The visibility of the multireddit, either ``"private"``,
                         ``"public"``, or ``"hidden"``.
    ==================== ==============================================================

    .. _unix time: https://en.wikipedia.org/wiki/Unix_time

    """

    STR_FIELD = "path"
    RE_INVALID = re.compile(r"[\W_]+", re.UNICODE)

    @staticmethod
    def sluggify(title: str):
        """Return a slug version of the title.

        :param title: The title to make a slug of.

        Adapted from Reddit's utils.py.

        """
        title = Multireddit.RE_INVALID.sub("_", title).strip("_").lower()
        if len(title) > 21:  # truncate to nearest word
            title = title[:21]
            last_word = title.rfind("_")
            if last_word > 0:
                title = title[:last_word]
        return title or "_"

    @cachedproperty
    def stream(self) -> SubredditStream:
        """Provide an instance of :class:`.SubredditStream`.

        Streams can be used to indefinitely retrieve new comments made to a multireddit,
        like:

        .. code-block:: python

            multireddit = await reddit.multireddit(redditor="spez", name="fun")
            async for comment in multireddit.stream.comments():
                print(comment)

        Additionally, new submissions can be retrieved via the stream. In the following
        example all new submissions to the multireddit are fetched:

        .. code-block:: python

            multireddit = await reddit.multireddit(redditor="bboe", name="games")
            async for submission in multireddit.stream.submissions():
                print(submission)

        """
        return SubredditStream(self)

    def __init__(self, reddit: "asyncpraw.Reddit", _data: Dict[str, Any]):
        """Initialize a :class:`.Multireddit` instance."""
        self.path = None
        super().__init__(reddit, _data=_data)
        self._author = Redditor(reddit, self.path.split("/", 3)[2])
        self._path = API_PATH["multireddit"].format(multi=self.name, user=self._author)
        self.path = f"/{self._path[:-1]}"  # Prevent requests for path
        if "subreddits" in self.__dict__:
            self.subreddits = [Subreddit(reddit, x["name"]) for x in self.subreddits]

    async def _ensure_author_fetched(self):
        if not self._author._fetched:
            await self._author._fetch()

    async def _fetch_info(self):
        await self._ensure_author_fetched()
        return (
            "multireddit_api",
            {"multi": self.name, "user": self._author.name},
            None,
        )

    async def _fetch_data(self):
        name, fields, params = await self._fetch_info()
        path = API_PATH[name].format(**fields)
        return await self._reddit.request(method="GET", params=params, path=path)

    async def _fetch(self):
        data = await self._fetch_data()
        data = data["data"]
        other = type(self)(self._reddit, _data=data)
        self.__dict__.update(other.__dict__)
        self._fetched = True

    async def add(self, subreddit: "asyncpraw.models.Subreddit"):
        """Add a subreddit to this multireddit.

        :param subreddit: The subreddit to add to this multi.

        For example, to add r/test to multireddit ``bboe/test``:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            multireddit = await reddit.multireddit(redditor="bboe", name="test")
            await multireddit.add(subreddit)

        """
        await self._ensure_author_fetched()
        url = API_PATH["multireddit_update"].format(
            multi=self.name, user=self._author, subreddit=subreddit
        )
        await self._reddit.put(url, data={"model": dumps({"name": str(subreddit)})})
        self._reset_attributes("subreddits")

    @_deprecate_args("display_name")
    async def copy(
        self, *, display_name: Optional[str] = None
    ) -> "asyncpraw.models.Multireddit":
        """Copy this multireddit and return the new multireddit.

        :param display_name: The display name for the copied multireddit. Reddit will
            generate the ``name`` field from this display name. When not provided the
            copy will use the same display name and name as this multireddit.

        To copy the multireddit ``bboe/test`` with a name of ``"testing"``:

        .. code-block:: python

            multireddit = await reddit.multireddit(redditor="bboe", name="test")
            await multireddit.copy(display_name="testing")

        """
        if display_name:
            name = self.sluggify(display_name)
        else:
            await self.load()
            display_name = self.display_name
            name = self.name
        data = {
            "display_name": display_name,
            "from": self.path,
            "to": API_PATH["multireddit"].format(
                multi=name, user=await self._reddit.user.me()
            ),
        }
        return await self._reddit.post(API_PATH["multireddit_copy"], data=data)

    async def delete(self):
        """Delete this multireddit.

        For example, to delete multireddit ``bboe/test``:

        .. code-block:: python

            multireddit = await reddit.multireddit(redditor="bboe", name="test")
            await multireddit.delete()

        """
        await self._ensure_author_fetched()
        path = API_PATH["multireddit_api"].format(
            multi=self.name, user=self._author.name
        )
        await self._reddit.delete(path)

    async def remove(self, subreddit: "asyncpraw.models.Subreddit"):
        """Remove a subreddit from this multireddit.

        :param subreddit: The subreddit to remove from this multi.

        For example, to remove r/test from multireddit ``bboe/test``:

        .. code-block:: python

            subreddit = await reddit.subreddit("test")
            multireddit = await reddit.multireddit(redditor="bboe", name="test")
            await multireddit.remove(subreddit)

        """
        await self._ensure_author_fetched()
        url = API_PATH["multireddit_update"].format(
            multi=self.name, user=self._author, subreddit=subreddit
        )
        await self._reddit.delete(url, data={"model": dumps({"name": str(subreddit)})})
        self._reset_attributes("subreddits")

    async def update(
        self,
        **updated_settings: Union[
            str, List[Union[str, "asyncpraw.models.Subreddit", Dict[str, str]]]
        ],
    ):
        """Update this multireddit.

        Keyword arguments are passed for settings that should be updated. They can any
        of:

        :param display_name: The display name for this multireddit. Must be no longer
            than 50 characters.
        :param subreddits: Subreddits for this multireddit.
        :param description_md: Description for this multireddit, formatted in Markdown.
        :param icon_name: Can be one of: ``"art and design"``, ``"ask"``, ``"books"``,
            ``"business"``, ``"cars"``, ``"comics"``, ``"cute animals"``, ``"diy"``,
            ``"entertainment"``, ``"food and drink"``, ``"funny"``, ``"games"``,
            ``"grooming"``, ``"health"``, ``"life advice"``, ``"military"``, ``"models
            pinup"``, ``"music"``, ``"news"``, ``"philosophy"``, ``"pictures and
            gifs"``, ``"science"``, ``"shopping"``, ``"sports"``, ``"style"``,
            ``"tech"``, ``"travel"``, ``"unusual stories"``, ``"video"``, or ``None``.
        :param key_color: RGB hex color code of the form ``"#FFFFFF"``.
        :param visibility: Can be one of: ``"hidden"``, ``"private"``, or ``"public"``.
        :param weighting_scheme: Can be one of: ``"classic"`` or ``"fresh"``.

        For example, to rename multireddit ``"bboe/test"`` to ``"bboe/testing"``:

        .. code-block:: python

            multireddit = await reddit.multireddit(redditor="bboe", name="test")
            await multireddit.update(display_name="testing")

        """
        if "subreddits" in updated_settings:
            updated_settings["subreddits"] = [
                {"name": str(sub)} for sub in updated_settings["subreddits"]
            ]
        await self._ensure_author_fetched()
        path = API_PATH["multireddit_api"].format(
            multi=self.name, user=self._author.name
        )
        new = await self._reddit.put(path, data={"model": dumps(updated_settings)})
        self.__dict__.update(new.__dict__)



"""Test asyncpraw.models.subreddits."""
import pytest

from .. import UnitTest


class TestSubreddits(UnitTest):
    async def test_recommended__invalid_omit_subreddits(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            await reddit.subreddits.recommended(["earthporn"], "invalid")
        assert str(excinfo.value) == "omit_subreddits must be a list or None"

    async def test_recommended__invalid_subreddits(self, reddit):
        with pytest.raises(TypeError) as excinfo:
            await reddit.subreddits.recommended("earthporn")
        assert str(excinfo.value) == "subreddits must be a list"

    def test_search__params_not_modified(self, reddit):
        params = {"dummy": "value"}
        generator = reddit.subreddits.search(None, params=params)
        assert generator.params["dummy"] == "value"
        assert params == {"dummy": "value"}



from asyncpraw.models import InlineGif, InlineImage, InlineMedia, InlineVideo

from ... import UnitTest


class TestInlineMedia(UnitTest):
    def test_equality(self):
        media1 = InlineMedia(path="path1", caption="caption1")
        media1.media_id = "media_id1"
        media2 = InlineMedia(path="path1", caption="caption1")
        media2.media_id = "media_id1"
        media3 = InlineMedia(path="path2", caption="caption2")
        media3.media_id = "media_id2"
        assert media1 == media1
        assert media2 == media2
        assert media3 == media3
        assert media1 == media2
        assert media2 != media3
        assert media1 != media3

    def test_repr(self):
        media = InlineMedia(path="path1", caption="caption1")
        no_caption = InlineMedia(path="path1")
        gif = InlineGif(path="gif_path1", caption="gif_caption1")
        image = InlineImage(path="image_path1", caption="image_caption1")
        video = InlineVideo(path="video_path1", caption="video_caption1")
        assert repr(media) == "<InlineMedia caption='caption1'>"
        assert repr(no_caption) == "<InlineMedia caption=None>"
        assert repr(gif) == "<InlineGif caption='gif_caption1'>"
        assert repr(image) == "<InlineImage caption='image_caption1'>"
        assert repr(video) == "<InlineVideo caption='video_caption1'>"

    def test_str(self):
        media = InlineMedia(path="path1", caption="caption1")
        no_caption = InlineMedia(path="path1")
        gif = InlineGif(path="gif_path1", caption="gif_caption1")
        image = InlineImage(path="image_path1", caption="image_caption1")
        video = InlineVideo(path="video_path1", caption="video_caption1")
        media.media_id = "media_media_id"
        no_caption.media_id = "media_media_id_no_caption"
        gif.media_id = "gif_media_id"
        image.media_id = "image_media_id"
        video.media_id = "video_media_id"
        assert str(media) == '\n\n![None](media_media_id "caption1")\n\n'
        assert str(no_caption) == '\n\n![None](media_media_id_no_caption "")\n\n'
        assert str(gif) == '\n\n![gif](gif_media_id "gif_caption1")\n\n'
        assert str(image) == '\n\n![img](image_media_id "image_caption1")\n\n'
        assert str(video) == '\n\n![video](video_media_id "video_caption1")\n\n'



import pytest

from asyncpraw import Reddit


@pytest.fixture
async def reddit():
    async with Reddit(
        client_id="dummy", client_secret="dummy", user_agent="dummy"
    ) as reddit:
        # Unit tests should never issue requests
        yield reddit



"""Test asyncpraw.models.listing.generator."""
import pytest

from asyncpraw.models.listing.generator import ListingGenerator

from ... import UnitTest


class TestListingGenerator(UnitTest):
    def test_params_are_not_modified(self):
        params = {"prawtest": "yes"}
        generator = ListingGenerator(None, None, params=params)
        assert "limit" in generator.params
        assert "limit" not in params
        assert ("prawtest", "yes") in generator.params.items()

    def test_bad_dict(self):
        generator = ListingGenerator(None, None)
        with pytest.raises(ValueError) as excinfo:
            generator._extract_sublist({"not_users": "test value"})
        assert excinfo.value.args[0] == (
            "The generator returned a dictionary Async PRAW didn't recognize. File a"
            " bug report at Async PRAW."
        )



import pytest

from asyncpraw.models import Rule, Subreddit

from ... import UnitTest


class TestRules(UnitTest):
    @property
    def subreddit(self):
        return Subreddit(None, display_name=pytest.placeholders.test_subreddit)

    def test_empty_value(self, reddit):
        with pytest.raises(ValueError):
            Rule(reddit, self.subreddit, short_name="")

    def test_no_data(self, reddit):
        with pytest.raises(ValueError) as excinfo:
            Rule(reddit, self.subreddit)
        assert excinfo.value.args[0] == "Either short_name or _data needs to be given."

    def test_both_data(self, reddit):
        with pytest.raises(ValueError) as excinfo:
            Rule(reddit, self.subreddit, short_name="test", _data={})
        assert excinfo.value.args[0] == "Either short_name or _data needs to be given."

    def test_no_subreddit(self, reddit):
        rule = Rule(reddit, short_name="test")
        with pytest.raises(ValueError) as excinfo:
            getattr(rule, "subreddit")
        assert (
            excinfo.value.args[0]
            == "The Rule is missing a subreddit. File a bug report at Async PRAW."
        )



"""This file should be updated as files/classes/functions are deprecated."""

import pytest

from asyncpraw import Reddit
from asyncpraw.exceptions import APIException, AsyncPRAWException, WebSocketException
from asyncpraw.models import Comment, Subreddit
from asyncpraw.models.reddit.user_subreddit import UserSubreddit
from asyncpraw.util.token_manager import FileTokenManager

from . import UnitTest

pytestmark = pytest.mark.filterwarnings(
    "ignore:Unclosed client session", category=ResourceWarning
)


@pytest.mark.filterwarnings("error", category=DeprecationWarning)
class TestDeprecation(UnitTest):
    def test_api_exception(self):
        exc = APIException(["test", "testing", "test"])
        with pytest.raises(DeprecationWarning):
            exc.error_type
        with pytest.raises(DeprecationWarning):
            exc.message
        with pytest.raises(DeprecationWarning):
            exc.field

    async def test_comment_forest_async_iterator(self, reddit):
        submission = await reddit.submission("1234", fetch=False)
        submission._fetched = True
        submission.comments._comments = [Comment(None, id="1234")]
        with pytest.deprecated_call():
            async for comment in submission.comments:
                assert isinstance(comment, Comment)

    async def test_comment_forest_list_async(self, reddit):
        submission = await reddit.submission("1234", fetch=False)
        submission._fetched = True
        submission.comments._comments = []
        with pytest.deprecated_call():
            await submission.comments.list()

    async def test_conversations_after_argument(self, reddit):
        with pytest.deprecated_call():
            subreddit = await reddit.subreddit("all")
            subreddit.modmail.conversations(after="after")

    async def test_gild_method(self, reddit):
        with pytest.raises(DeprecationWarning) as excinfo:
            submission = await reddit.submission("1234", fetch=False)
            await submission.gild()
            assert excinfo.value.args[0] == "'.gild' has been renamed to '.award'."

    def test_gold_method(self, reddit):
        with pytest.raises(DeprecationWarning) as excinfo:
            reddit.subreddits.gold()
            assert (
                excinfo.value.args[0]
                == "'subreddits.gold' has be renamed to 'subreddits.premium'."
            )

    async def test_lazy_argument_rename(self, reddit):
        with pytest.deprecated_call() as warning_info:
            await reddit.submission("1234", lazy=True)
        assert (
            str(warning_info.list[0].message)
            == "The parameter ``lazy`` has been renamed to ``fetch`` and support for"
            " the ``lazy`` parameter will be removed in a future version of Async PRAW."
        )

    def test_praw_exception_rename(self):
        with pytest.raises(AsyncPRAWException):
            Reddit()

        with pytest.raises(DeprecationWarning):
            import asyncpraw

            asyncpraw.exceptions.PRAWException

        with pytest.raises(DeprecationWarning):
            from asyncpraw import exceptions

            exceptions.PRAWException

        with pytest.raises(DeprecationWarning):
            from asyncpraw.exceptions import PRAWException  # noqa: F401

    async def test_reddit_token_manager(self):
        with pytest.raises(DeprecationWarning):
            async with Reddit(
                token_manager=FileTokenManager("name"),
                client_id="dummy",
                client_secret=None,
                redirect_uri="dummy",
                user_agent="dummy",
            ) as reddit:
                reddit._core._requestor._http = None

    async def test_reddit_user_me_read_only(self, reddit):
        with pytest.raises(DeprecationWarning):
            await reddit.user.me()

    async def test_submission_comments_async(self, reddit):
        submission = await reddit.submission("1234", fetch=False)
        submission._fetched = True
        with pytest.deprecated_call():
            await submission.comments()

    async def test_subreddit_rules_call(self, reddit):
        with pytest.raises(DeprecationWarning) as excinfo:
            subreddit = Subreddit(reddit, display_name="test")
            await subreddit.rules()
        assert (
            excinfo.value.args[0]
            == "Calling SubredditRules to get a list of rules is deprecated. Remove the"
            " parentheses to use the iterator. View the Async PRAW documentation on how"
            " to change the code in order to use the iterator"
            " (https://asyncpraw.readthedocs.io/en/latest/code_overview/other/subredditrules.html#asyncpraw.models.reddit.rules.SubredditRules.__call__)."
        )

    async def test_synchronous_context_manager(self, reddit):
        with pytest.raises(DeprecationWarning) as excinfo:
            with reddit:
                pass
            assert (
                excinfo.value.args[0]
                == "Using this class as a synchronous context manager is deprecated and"
                " will be removed in the next release. Use this class as an"
                " asynchronous context manager instead."
            )

    def test_user_subreddit_as_dict(self):
        user_subreddit = UserSubreddit(None, display_name="test")
        with pytest.deprecated_call() as warning_info:
            display_name = user_subreddit["display_name"]
            assert display_name == "test"
            assert (
                warning_info.list[0].message.args[0]
                == "'Redditor.subreddit' is no longer a dict and is now an"
                " UserSubreddit object. Accessing attributes using string indices is"
                " deprecated."
            )
            assert user_subreddit.keys() == user_subreddit.__dict__.keys()
            assert (
                warning_info.list[1].message.args[0]
                == "'Redditor.subreddit' is no longer a dict and is now an"
                " UserSubreddit object. Using 'keys' is deprecated and will be removed"
                " in Async PRAW 8."
            )

    def test_validate_on_submit(self, reddit):
        with pytest.raises(DeprecationWarning):
            reddit.validate_on_submit
        reddit.validate_on_submit = True
        assert reddit.validate_on_submit
        reddit.validate_on_submit = False
        with pytest.raises(DeprecationWarning):
            reddit.validate_on_submit

    def test_web_socket_exception_attribute(self):
        exc = WebSocketException("Test", Exception("Test"))
        with pytest.raises(DeprecationWarning) as excinfo:
            _ = exc.original_exception
        assert (
            excinfo.value.args[0]
            == "Accessing the attribute 'original_exception' is deprecated. Please"
            " rewrite your code in such a way that this attribute does not need to"
            " be used. It will be removed in Async PRAW 8.0."
        )



# coding: utf-8
import pytest

from asyncpraw.exceptions import (
    APIException,
    AsyncPRAWException,
    ClientException,
    DuplicateReplaceException,
    InvalidFlairTemplateID,
    InvalidImplicitAuth,
    InvalidURL,
    MediaPostFailed,
    MissingRequiredAttributeException,
    RedditAPIException,
    RedditErrorItem,
    WebSocketException,
)


class TestAsyncPRAWException:
    def test_inheritance(self):
        assert issubclass(AsyncPRAWException, Exception)

    def test_str(self):
        assert str(AsyncPRAWException()) == ""
        assert str(AsyncPRAWException("foo")) == "foo"


class TestRedditErrorItem:
    def test_equality(self):
        resp = {
            "error_type": "BAD_SOMETHING",
            "field": "some_field",
            "message": "invalid something",
        }
        error = RedditErrorItem(**resp)
        error2 = RedditErrorItem(**resp)
        assert error == error2
        assert error != 0

    def test_property(self):
        error = RedditErrorItem(
            "BAD_SOMETHING", field="some_field", message="invalid something"
        )
        assert (
            error.error_message
            == "BAD_SOMETHING: 'invalid something' on field 'some_field'"
        )

    def test_str(self):
        error = RedditErrorItem(
            "BAD_SOMETHING", field="some_field", message="invalid something"
        )
        assert str(error) == "BAD_SOMETHING: 'invalid something' on field 'some_field'"

    def test_repr(self):
        error = RedditErrorItem(
            "BAD_SOMETHING", field="some_field", message="invalid something"
        )
        assert (
            repr(error)
            == "RedditErrorItem(error_type='BAD_SOMETHING', message='invalid"
            " something', field='some_field')"
        )


class TestAPIException:
    def test_catch(self):
        exc = RedditAPIException([["test", "testing", "test"]])
        with pytest.raises(APIException):
            raise exc


class TestRedditAPIException:
    def test_inheritance(self):
        assert issubclass(RedditAPIException, AsyncPRAWException)

    def test_items(self):
        container = RedditAPIException(
            [
                ["BAD_SOMETHING", "invalid something", "some_field"],
                RedditErrorItem(
                    "BAD_SOMETHING", field="some_field", message="invalid something"
                ),
            ]
        )
        for exception in container.items:
            assert isinstance(exception, RedditErrorItem)

    @pytest.mark.filterwarnings("ignore", category=DeprecationWarning)
    def test_apiexception_value(self):
        exc = RedditAPIException("test", "testing", "test")
        assert exc.error_type == "test"
        exc2 = RedditAPIException(["test", "testing", "test"])
        assert exc2.message == "testing"
        exc3 = RedditAPIException([["test", "testing", "test"]])
        assert exc3.field == "test"


class TestClientException:
    def test_inheritance(self):
        assert issubclass(ClientException, AsyncPRAWException)

    def test_str(self):
        assert str(ClientException()) == ""
        assert str(ClientException("error message")) == "error message"


class TestDuplicateReplaceException:
    def test_inheritance(self):
        assert issubclass(DuplicateReplaceException, ClientException)

    def test_message(self):
        assert (
            str(DuplicateReplaceException())
            == "A duplicate comment has been detected. Are you attempting to call"
            " 'replace_more_comments' more than once?"
        )


class TestInvalidFlairTemplateID:
    def test_inheritance(self):
        assert issubclass(InvalidFlairTemplateID, ClientException)

    def test_str(self):
        assert (
            str(InvalidFlairTemplateID("123"))
            == "The flair template ID '123' is invalid. If you are trying to create a"
            " flair, please use the 'add' method."
        )


class TestInvalidImplicitAuth:
    def test_inheritance(self):
        assert issubclass(InvalidImplicitAuth, ClientException)

    def test_message(self):
        assert (
            str(InvalidImplicitAuth())
            == "Implicit authorization can only be used with installed apps."
        )


class TestInvalidURL:
    def test_inheritance(self):
        assert issubclass(InvalidURL, ClientException)

    def test_message(self):
        assert (
            str(InvalidURL("https://www.google.com"))
            == "Invalid URL: https://www.google.com"
        )

    def test_custom_message(self):
        assert (
            str(InvalidURL("https://www.google.com", message="Test custom {}"))
            == "Test custom https://www.google.com"
        )


class TestMissingRequiredAttributeException:
    def test_inheritance(self):
        assert issubclass(MissingRequiredAttributeException, ClientException)

    def test_str(self):
        assert str(MissingRequiredAttributeException()) == ""
        assert (
            str(MissingRequiredAttributeException("error message")) == "error message"
        )


class TestWebSocketException:
    def test_inheritance(self):
        assert issubclass(WebSocketException, ClientException)

    def test_str(self):
        assert str(WebSocketException("", None)) == ""
        assert str(WebSocketException("error message", None)) == "error message"

    @pytest.mark.filterwarnings("ignore", category=DeprecationWarning)
    def test_exception_attr(self):
        exc = WebSocketException(None, None)
        assert exc.original_exception is None
        assert isinstance(WebSocketException(None, Exception()), Exception)
        assert (
            str(WebSocketException(None, Exception("test")).original_exception)
            == "test"
        )
        exc.original_exception = Exception()
        assert isinstance(exc.original_exception, Exception)
        del exc.original_exception
        assert "_original_exception" not in vars(exc)


class TestMediaPostFailed:
    def test_inheritance(self):
        assert issubclass(MediaPostFailed, WebSocketException)

    def test_message(self):
        assert (
            str(MediaPostFailed())
            == "The attempted media upload action has failed. Possible causes include"
            " the corruption of media files. Check that the media file can be opened"
            " on your local machine."
        )



import pytest

from asyncpraw.exceptions import ClientException
from asyncpraw.models import Comment

from ... import UnitTest


class TestComment(UnitTest):
    def test_attribute_error(self, reddit):
        with pytest.raises(AttributeError):
            Comment(reddit, _data={"id": "1"}).mark_as_read()

    def test_equality(self, reddit):
        comment1 = Comment(reddit, _data={"id": "dummy1", "n": 1})
        comment2 = Comment(reddit, _data={"id": "Dummy1", "n": 2})
        comment3 = Comment(reddit, _data={"id": "dummy3", "n": 2})
        assert comment1 == comment1
        assert comment2 == comment2
        assert comment3 == comment3
        assert comment1 == comment2
        assert comment2 != comment3
        assert comment1 != comment3
        assert "dummy1" == comment1
        assert comment2 == "dummy1"

    def test_construct_failure(self, reddit):
        message = "Exactly one of 'id', 'url', or '_data' must be provided."
        with pytest.raises(TypeError) as excinfo:
            Comment(reddit)
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Comment(reddit, id="dummy", url="dummy")
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Comment(reddit, "dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Comment(reddit, url="dummy", _data={"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(TypeError) as excinfo:
            Comment(reddit, "dummy", "dummy", {"id": "dummy"})
        assert str(excinfo.value) == message

        with pytest.raises(ValueError):
            Comment(reddit, "")
        with pytest.raises(ValueError):
            Comment(reddit, url="")

    def test_construct_from_url(self, reddit):
        url = "https://reddit.com/comments/2gmzqe/_/cklhv0f/"
        assert Comment(reddit, url=url) == "cklhv0f"

    def test_hash(self, reddit):
        comment1 = Comment(reddit, _data={"id": "dummy1", "n": 1})
        comment2 = Comment(reddit, _data={"id": "Dummy1", "n": 2})
        comment3 = Comment(reddit, _data={"id": "dummy3", "n": 2})
        assert hash(comment1) == hash(comment1)
        assert hash(comment2) == hash(comment2)
        assert hash(comment3) == hash(comment3)
        assert hash(comment1) == hash(comment2)
        assert hash(comment2) != hash(comment3)
        assert hash(comment1) != hash(comment3)

    def test_id_from_url(self):
        urls = [
            "http://reddit.com/comments/2gmzqe/_/cklhv0f/",
            "https://reddit.com/comments/2gmzqe/_/cklhv0f",
            "http://www.reddit.com/r/redditdev/comments/2gmzqe/_/cklhv0f/",
            "https://www.reddit.com/r/redditdev/comments/2gmzqe/_/cklhv0f",
        ]
        for url in urls:
            assert Comment.id_from_url(url) == "cklhv0f", url

    def test_id_from_url__invalid_urls(self):
        urls = [
            "",
            "1",
            "/",
            "my.it/2gmzqe",
            "http://my.it/_",
            "https://redd.it/_/",
            "http://reddit.com/comments/_/2gmzqe",
            "http://my.it/2gmzqe",
            "https://redd.it/2gmzqe",
            "http://reddit.com/comments/2gmzqe",
            "https://www.reddit.com/r/redditdev/comments/2gmzqe/",
        ]
        for url in urls:
            with pytest.raises(ClientException):
                Comment.id_from_url(url)

    def test_repr(self, reddit):
        comment = Comment(reddit, id="dummy")
        assert repr(comment) == "Comment(id='dummy')"

    def test_str(self, reddit):
        comment = Comment(reddit, _data={"id": "dummy"})
        assert str(comment) == "dummy"

    def test_unset_hidden_attribute_does_not_fetch(self, reddit):
        comment = Comment(reddit, _data={"id": "dummy"})
        assert comment._fetched
        with pytest.raises(AttributeError):
            comment._ipython_canary_method_should_not_exist_



import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import ClientException
from asyncpraw.models import RemovalReason

from ... import IntegrationTest


class TestRemovalReason(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test__fetch(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            reason = await subreddit.mod.removal_reasons.get_reason("159bqhvme3rxe")
            assert reason.title.startswith("Be Kind")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test__fetch_int(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette("TestRemovalReason.test__fetch"):
            reason = await subreddit.mod.removal_reasons.get_reason(0)
            assert isinstance(reason, RemovalReason)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test__fetch_slice(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette("TestRemovalReason.test__fetch"):
            reasons = await subreddit.mod.removal_reasons.get_reason(slice(-3, None))
            assert len(reasons) == 3
            for reason in reasons:
                assert isinstance(reason, RemovalReason)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test__fetch__invalid_reason(self, _):
        self.reddit.read_only = False
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            with pytest.raises(ClientException) as excinfo:
                await subreddit.mod.removal_reasons.get_reason("invalid")
            assert (
                str(excinfo.value)
                == f"Subreddit {subreddit} does not have the removal reason invalid"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            reason = await subreddit.mod.removal_reasons.get_reason("159bqhvme3rxe")
            await reason.update(title="New Title", message="New Message")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_empty(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            reasons = [reason async for reason in subreddit.mod.removal_reasons]
            reason = reasons[0]
            await reason.update()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            reason = await subreddit.mod.removal_reasons.get_reason("157l8fono55wf")
            await reason.delete()


class TestSubredditRemovalReasons(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test__aiter(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            count = 0
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            async for reason in subreddit.mod.removal_reasons:
                assert isinstance(reason, RemovalReason)
                count += 1
            assert count > 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            reason = await subreddit.mod.removal_reasons.add(
                title="Test", message="test"
            )
            assert isinstance(reason, RemovalReason)



import sys

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import Preferences

from .. import IntegrationTest


class TestPreferences(IntegrationTest):
    def test_creation(self):
        prefs_obj = self.reddit.user.preferences
        assert isinstance(prefs_obj, Preferences)

    async def test_view(self):
        some_known_keys = {
            "allow_clicktracking",
            "default_comment_sort",
            "hide_from_robots",
            "lang",
            "no_profanity",
            "over_18",
            "public_votes",
            "show_link_flair",
        }

        self.reddit.read_only = False
        with self.use_cassette():
            prefs_dict = await self.reddit.user.preferences()
        assert isinstance(prefs_dict, dict)
        assert some_known_keys.issubset(prefs_dict)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        # boolean params, as many as are reproducible on multiple accounts.
        bool_params = (
            "allow_clicktracking",
            "beta",
            "clickgadget",
            "collapse_read_messages",
            "compress",
            "domain_details",
            "email_digests",
            "email_messages",
            "email_unsubscribe_all",
            "enable_default_themes",
            "hide_downs",
            "hide_from_robots",
            "hide_ups",
            "highlight_controversial",
            "highlight_new_comments",
            "ignore_suggested_sort",
            "legacy_search",
            "live_orangereds",
            "mark_messages_read",
            "monitor_mentions",
            "newwindow",
            "over_18",
            "private_feeds",
            "profile_opt_out",
            "public_votes",
            "research",
            "search_include_over_18",
            "show_flair",
            "show_link_flair",
            "show_stylesheets",
            "show_trending",
            "store_visits",
            "threaded_messages",
            "threaded_modmail",
            "top_karma_subreddits",
            "use_global_defaults",
        )

        int_params = (
            "min_comment_score",
            "min_link_score",
            "num_comments",
            "numsites",
        )
        # parameters that accept string types, and two valid values
        str_params = (
            ("default_comment_sort", ("confidence", "top")),
            ("lang", ("es", "en")),
            ("media", ("on", "off")),
            ("media_preview", ("on", "off")),
        )

        # there are also subreddit-name parameters related to the Gold
        # styling feature. It's impractical to test for that because not every
        # account has Gold, and the test fails on normal accounts.

        self.reddit.read_only = False
        preferences = self.reddit.user.preferences

        with self.use_cassette():

            # test an empty update
            await preferences.update()

            for param in int_params:
                response = await preferences.update(**{param: 1})
                assert response[param] == 1
                response = await preferences.update(**{param: 3})
                assert response[param] == 3
            for param in bool_params:
                response = await preferences.update(**{param: True})
                assert response[param] is True
                response = await preferences.update(**{param: False})
                assert response[param] is False
            for param, values in str_params:
                response = await preferences.update(**{param: values[0]})
                assert response[param] == values[0]
                response = await preferences.update(**{param: values[1]})
                assert response[param] == values[1]



from asyncpraw.util import camel_to_snake, snake_case_keys

from .. import UnitTest


class TestSnake(UnitTest):
    def test_camel_to_snake(self):
        test_strings = (
            ("camelCase", "camel_case"),
            ("PascalCase", "pascal_case"),
            ("camelCasePlace", "camel_case_place"),
            ("Pascal8Camel8Snake", "pascal8_camel8_snake"),
            ("HTTPResponseCode", "http_response_code"),
            ("ResponseHTTP", "response_http"),
            ("ResponseHTTP200", "response_http200"),
            ("getHTTPResponseCode", "get_http_response_code"),
            ("get200HTTPResponseCode", "get200_http_response_code"),
            ("getHTTP200ResponseCode", "get_http200_response_code"),
            ("12PolarBears", "12_polar_bears"),
            ("11buzzingBees", "11buzzing_bees"),
            ("TacocaT", "tacoca_t"),
            ("fooBARbaz", "foo_ba_rbaz"),
            ("foo_BAR_baz", "foo_bar_baz"),
            ("Base_BASE", "base_base"),
            ("Case_Case", "case_case"),
            ("FACE_Face", "face_face"),
        )
        for camel, snake in test_strings:
            assert camel_to_snake(camel) == snake

    def test_camel_to_snake_dict(self):
        test_strings = {
            "camelCase": "camel_case",
            "PascalCase": "pascal_case",
            "camelCasePlace": "camel_case_place",
            "Pascal8Camel8Snake": "pascal8_camel8_snake",
            "HTTPResponseCode": "http_response_code",
            "ResponseHTTP": "response_http",
            "ResponseHTTP200": "response_http200",
            "getHTTPResponseCode": "get_http_response_code",
            "get200HTTPResponseCode": "get200_http_response_code",
            "getHTTP200ResponseCode": "get_http200_response_code",
            "12PolarBears": "12_polar_bears",
            "11buzzingBees": "11buzzing_bees",
            "TacocaT": "tacoca_t",
            "fooBARbaz": "foo_ba_rbaz",
            "foo_BAR_baz": "foo_bar_baz",
            "Base_BASE": "base_base",
            "Case_Case": "case_case",
            "FACE_Face": "face_face",
        }
        new_test = snake_case_keys(test_strings)
        for key, item in new_test.items():
            assert key == item



import sys

import pytest

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import RedditAPIException
from asyncpraw.models import Comment, Submission

from ... import IntegrationTest


class TestSubmission(IntegrationTest):
    async def test_comments(self):
        with self.use_cassette():
            submission = await self.reddit.submission("2gmzqe")
            assert len(submission.comments) == 1
            assert isinstance(submission.comments[0], Comment)
            assert isinstance(submission.comments[0].replies[0], Comment)

    async def test_comments__fetch_async_call(self):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("2gmzqe", fetch=False)
            with pytest.deprecated_call():
                await submission.comments()
                assert submission._fetched
                assert submission.comments

    async def test_clear_vote(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").clear_vote()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkbt8")
            await submission.delete()
            await submission.load()
            assert submission.author is None
            assert submission.selftext == "[deleted]"

    async def test_disable_inbox_replies(self):
        self.reddit.read_only = False
        submission = Submission(self.reddit, "hmkbt8")
        with self.use_cassette():
            await submission.disable_inbox_replies()

    async def test_downvote(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").downvote()

    async def test_duplicates(self):
        with self.use_cassette():
            submission = Submission(self.reddit, "avj2v")
            assert len(await self.async_list(submission.duplicates())) > 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_edit(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkbt8")
            await submission.edit("New text")
            assert submission.selftext == "New text"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_edit_invalid(self, _):
        self.reddit.read_only = False
        self.reddit.validate_on_submit = True
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkfoy")
            with pytest.raises(RedditAPIException):
                await submission.edit("rewtwert")

    async def test_enable_inbox_replies(self):
        self.reddit.read_only = False
        submission = Submission(self.reddit, "hmkbt8")
        with self.use_cassette():
            await submission.enable_inbox_replies()

    async def test_award(self):
        self.reddit.read_only = False
        with self.use_cassette():
            award_data = await Submission(self.reddit, "j3kyoo").award()
            assert award_data["gildings"]["gid_2"] == 2

    async def test_award__not_enough_coins(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await Submission(self.reddit, "j3kyoo").award(
                    gild_type="award_2385c499-a1fb-44ec-b9b7-d260f3dc55de"
                )
            exception = excinfo.value
            assert "INSUFFICIENT_COINS_WITH_AMOUNT" == exception.error_type

    async def test_award__self_gild(self):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await Submission(self.reddit, "j3fkiw").award(
                    gild_type="award_2385c499-a1fb-44ec-b9b7-d260f3dc55de"
                )
            exception = excinfo.value
            assert "SELF_GILDING_NOT_ALLOWED" == exception.error_type

    async def test_gild(self):
        self.reddit.read_only = False
        with self.use_cassette("TestSubmission.test_award"):
            award_data = await Submission(self.reddit, "j3kyoo").gild()
            assert award_data["gildings"]["gid_2"] == 2

    async def test_gilded(self):
        with self.use_cassette():
            submission = await self.reddit.submission("2gmzqe")
            assert 1 == submission.gilded

    async def test_hide(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").hide()

    async def test_hide_multiple(self):
        self.reddit.read_only = False
        submissions = [
            Submission(self.reddit, "fewoh"),
            Submission(self.reddit, "c625v"),
        ]
        with self.use_cassette():
            await Submission(self.reddit, "1eipl7").hide(other_submissions=submissions)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_hide_multiple_in_batches(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("popular")
            submissions = await self.async_list(subreddit.hot(limit=100))
            assert len(submissions) == 100
            await submissions[0].hide(other_submissions=submissions[1:])

    async def test_invalid_attribute(self):
        with self.use_cassette():
            with pytest.raises(AttributeError) as excinfo:
                submission = await self.reddit.submission("2gmzqe")
                submission.invalid_attribute
        assert (
            excinfo.value.args[0]
            == "'Submission' object has no attribute 'invalid_attribute'"
        )

    async def test_reply(self):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkbt8")
            comment = await submission.reply("Test reply")
            assert comment.author == pytest.placeholders.username
            assert comment.body == "Test reply"
            assert comment.parent_id == submission.fullname

    # async def test_reply__none(self): # TODO: I have not been able to reproduce this again; same with praw
    #     self.reddit.read_only = False
    #     submission = Submission(self.reddit, "ah19vv")
    #     with self.use_cassette():
    #         reply = submission.reply("TEST")
    #     assert reply is None

    async def test_report(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").report("praw")

    async def test_save(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").save()

    async def test_mark_visited(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkfoi").mark_visited()

    async def test_unhide(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").unhide()

    async def test_unhide_multiple(self):
        self.reddit.read_only = False
        submissions = [
            Submission(self.reddit, "fewoh"),
            Submission(self.reddit, "c625v"),
        ]
        with self.use_cassette():
            await Submission(self.reddit, "1eipl7").unhide(
                other_submissions=submissions
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unhide_multiple_in_batches(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("popular")
            submissions = await self.async_list(subreddit.hot(limit=100))
            assert len(submissions) == 100
            await submissions[0].unhide(other_submissions=submissions[1:])

    async def test_unsave(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").unsave()

    async def test_upvote(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").upvote()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_crosspost(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = pytest.placeholders.test_subreddit
            crosspost_parent = await self.reddit.submission("6vx01b")

            submission = await crosspost_parent.crosspost(subreddit)
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert submission.title == "Test Title"
            assert submission.crosspost_parent == "t3_6vx01b"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_crosspost__subreddit_object(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            crosspost_parent = await self.reddit.submission("6vx01b")

            submission = await crosspost_parent.crosspost(subreddit)
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert submission.title == "Test Title"
            assert submission.crosspost_parent == "t3_6vx01b"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_crosspost__custom_title(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
            crosspost_parent = await self.reddit.submission("6vx01b")

            submission = await crosspost_parent.crosspost(subreddit, title="my title")
            await submission.load()
            assert submission.author == pytest.placeholders.username
            assert submission.title == "my title"
            assert submission.crosspost_parent == "t3_6vx01b"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_crosspost__flair(self, _):
        flair_id = "94f13282-e2e8-11e8-8291-0eae4e167256"
        flair_text = "AF"
        flair_class = ""
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = pytest.placeholders.test_subreddit
            crosspost_parent = await self.reddit.submission("6vx01b")

            submission = await crosspost_parent.crosspost(
                subreddit, flair_id=flair_id, flair_text=flair_text
            )
            await submission.load()
            assert submission.link_flair_css_class == flair_class
            assert submission.link_flair_text == flair_text
            assert submission.crosspost_parent == "t3_6vx01b"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_crosspost__nsfw(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = pytest.placeholders.test_subreddit
            crosspost_parent = await self.reddit.submission("6vx01b")

            submission = await crosspost_parent.crosspost(subreddit, nsfw=True)
            await submission.load()
            assert submission.over_18 is True
            assert submission.crosspost_parent == "t3_6vx01b"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_crosspost__spoiler(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            subreddit = pytest.placeholders.test_subreddit
            crosspost_parent = await self.reddit.submission("6vx01b")

            submission = await crosspost_parent.crosspost(subreddit, spoiler=True)
            await submission.load()
            assert submission.spoiler is True
            assert submission.crosspost_parent == "t3_6vx01b"


class TestSubmissionFlair(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_choices(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkbt8")
            expected = [
                {
                    "flair_css_class": "",
                    "flair_position": "right",
                    "flair_template_id": "94f13282-e2e8-11e8-8291-0eae4e167256",
                    "flair_text": "AF",
                    "flair_text_editable": True,
                }
            ]
            choices = await submission.flair.choices()
            assert expected == choices

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_select(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkbt8")
            await submission.flair.select("94f13282-e2e8-11e8-8291-0eae4e167256")


class TestSubmissionModeration(IntegrationTest):
    async def test_approve(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.approve()

    async def test_contest_mode(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.contest_mode()

    async def test_contest_mode__disable(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.contest_mode(state=False)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_flair(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.flair(text="AF")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_flair_template_id(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.flair(
                text="submission flair",
                flair_template_id="94f13282-e2e8-11e8-8291-0eae4e167256",
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_flair_text_only(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.flair(text="submission flair")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_flair_text_and_css_class(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.flair(
                text="submission flair", css_class="submission flair"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_flair_all(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.flair(
                text="submission flair",
                css_class="submission flair",
                flair_template_id="94f13282-e2e8-11e8-8291-0eae4e167256",
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_flair_just_template_id(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.flair(
                flair_template_id="94f13282-e2e8-11e8-8291-0eae4e167256"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_flair_just_css_class(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.flair(
                css_class="submission flair"
            )

    async def test_distinguish(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hrzz2x").mod.distinguish()

    async def test_ignore_reports(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.ignore_reports()

    async def test_notes(self):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("uflrmv")
            note = await submission.mod.create_note(
                label="HELPFUL_USER", note="test note"
            )
            notes = await self.async_list(submission.mod.author_notes())
            assert note in notes
            assert notes[notes.index(note)].user == submission.author

    async def test_nsfw(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.nsfw()

    async def test_lock(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.lock()

    async def test_remove(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.remove(spam=True)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_remove_with_reason_id(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.remove(
                reason_id="159bqhvme3rxe"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_removal_reason(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkbt8")
            await submission.mod.remove()
            await submission.mod._add_removal_reason(
                mod_note="Foobar", reason_id="159bqhvme3rxe"
            )

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_removal_reason_without_id(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = Submission(self.reddit, "hmkbt8")
            await submission.mod.remove()
            await submission.mod._add_removal_reason(mod_note="Foobar")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add_removal_reason_without_id_or_note(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(ValueError) as excinfo:
                submission = Submission(self.reddit, "hmkbt8")
                await submission.mod.remove()
                await submission.mod._add_removal_reason()
            assert excinfo.value.args[0].startswith("mod_note cannot be blank")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_send_removal_message(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("hmkutx")
            mod = submission.mod
            await mod.remove()
            message = "message"
            res = [
                await mod.send_removal_message(
                    message=type, title="title", type=message
                )
                for type in ("public", "private", "private_exposed")
            ]
            assert isinstance(res[0], Comment)
            assert res[0].parent_id == f"t3_{submission.id}"
            assert res[0].stickied
            assert res[0].body == message
            assert res[1] is None
            assert res[2] is None

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_set_original_content(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("hmkbt8")
            assert not submission.is_original_content
            await submission.mod.set_original_content()
            submission = await self.reddit.submission("hmkbt8")
            assert submission.is_original_content

    async def test_sfw(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.sfw()

    async def test_spoiler(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.spoiler()

    async def test_sticky(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.sticky()

    async def test_sticky__remove(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.sticky(state=False)

    async def test_sticky__top(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.sticky(bottom=False)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_sticky__ignore_conflicts(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.sticky(bottom=False)
            await Submission(self.reddit, "hmkbt8").mod.sticky(bottom=False)

    async def test_suggested_sort(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.suggested_sort(sort="random")

    async def test_suggested_sort__clear(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.suggested_sort(sort="blank")

    async def test_undistinguish(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hrzz2x").mod.undistinguish()

    async def test_unignore_reports(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.unignore_reports()

    async def test_unlock(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.unlock()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unset_original_content(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("hmkbt8")
            assert submission.is_original_content
            await submission.mod.unset_original_content()
            submission = await self.reddit.submission("hmkbt8")
            assert not submission.is_original_content

    async def test_unspoiler(self):
        self.reddit.read_only = False
        with self.use_cassette():
            await Submission(self.reddit, "hmkbt8").mod.unspoiler()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update_crowd_control_level(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            submission = await self.reddit.submission("ol4d5w")
            await submission.mod.update_crowd_control_level(2)
            modlog = await self.async_next(
                submission.subreddit.mod.log(
                    action="adjust_post_crowd_control_level", limit=1
                )
            )
            assert modlog.action == "adjust_post_crowd_control_level"
            assert modlog.details == "medium"
            assert modlog.target_fullname == submission.fullname



"""Provide the RedditorListingMixin class."""
from typing import TYPE_CHECKING, AsyncIterator, Dict, Union
from urllib.parse import urljoin

from ....util.cache import cachedproperty
from ..generator import ListingGenerator
from .base import BaseListingMixin
from .gilded import GildedListingMixin

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class SubListing(BaseListingMixin):
    """Helper class for generating :class:`.ListingGenerator` objects."""

    def __init__(self, reddit: "asyncpraw.Reddit", base_path: str, subpath: str):
        """Initialize a :class:`.SubListing` instance.

        :param reddit: An instance of :class:`.Reddit`.
        :param base_path: The path to the object up to this point.
        :param subpath: The additional path to this sublisting.

        """
        super().__init__(reddit, _data=None)
        self._listing_use_sort = True
        self._reddit = reddit
        self._path = urljoin(base_path, subpath)


class RedditorListingMixin(BaseListingMixin, GildedListingMixin):
    """Adds additional methods pertaining to :class:`.Redditor` instances."""

    @cachedproperty
    def comments(self) -> SubListing:
        r"""Provide an instance of :class:`.SubListing` for comment access.

        For example, to output the first line of all new comments by u/spez try:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for comment in redditor.comments.new(limit=None):
                print(comment.body.split("\\n", 1)[0][:79])

        """
        return SubListing(self._reddit, self._path, "comments")

    @cachedproperty
    def submissions(self) -> SubListing:
        """Provide an instance of :class:`.SubListing` for submission access.

        For example, to output the title's of top 100 of all time submissions for u/spez
        try:

        .. code-block:: python

            redditor = await reddit.redditor("spez")
            async for submission in redditor.submissions.top(time_filter="all"):
                print(submission.title)

        """
        return SubListing(self._reddit, self._path, "submitted")

    def downvoted(
        self, **generator_kwargs: Union[str, int, Dict[str, str]]
    ) -> AsyncIterator[
        Union["asyncpraw.models.Comment", "asyncpraw.models.Submission"]
    ]:
        """Return a :class:`.ListingGenerator` for items the user has downvoted.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has downvoted.

        :raises: ``asyncprawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all downvoted items of the authenticated user:

        .. code-block:: python

            current_user = await reddit.user.me()
            async for item in current_user.downvoted():
                print(item.id)

        """
        return ListingGenerator(
            self._reddit, urljoin(self._path, "downvoted"), **generator_kwargs
        )

    def gildings(
        self, **generator_kwargs: Union[str, int, Dict[str, str]]
    ) -> AsyncIterator[
        Union["asyncpraw.models.Comment", "asyncpraw.models.Submission"]
    ]:
        """Return a :class:`.ListingGenerator` for items the user has gilded.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has gilded.

        :raises: ``asyncprawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all gilded items of the authenticated user:

        .. code-block:: python

            current_user = await reddit.user.me()
            async for item in current_user.gildings():
                print(item.id)

        """
        return ListingGenerator(
            self._reddit, urljoin(self._path, "gilded/given"), **generator_kwargs
        )

    def hidden(
        self, **generator_kwargs: Union[str, int, Dict[str, str]]
    ) -> AsyncIterator[
        Union["asyncpraw.models.Comment", "asyncpraw.models.Submission"]
    ]:
        """Return a :class:`.ListingGenerator` for items the user has hidden.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has hid.

        :raises: ``asyncprawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all hidden items of the authenticated user:

        .. code-block:: python

            current_user = await reddit.user.me()
            async for item in current_user.hidden():
                print(item.id)

        """
        return ListingGenerator(
            self._reddit, urljoin(self._path, "hidden"), **generator_kwargs
        )

    def saved(
        self, **generator_kwargs: Union[str, int, Dict[str, str]]
    ) -> AsyncIterator[
        Union["asyncpraw.models.Comment", "asyncpraw.models.Submission"]
    ]:
        """Return a :class:`.ListingGenerator` for items the user has saved.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has saved.

        :raises: ``asyncprawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all saved items of the authenticated user:

        .. code-block:: python

            current_user = await reddit.user.me()
            async for item in current_user.saved(limit=None):
                print(item.id)

        """
        return ListingGenerator(
            self._reddit, urljoin(self._path, "saved"), **generator_kwargs
        )

    def upvoted(
        self, **generator_kwargs: Union[str, int, Dict[str, str]]
    ) -> AsyncIterator[
        Union["asyncpraw.models.Comment", "asyncpraw.models.Submission"]
    ]:
        """Return a :class:`.ListingGenerator` for items the user has upvoted.

        :returns: A :class:`.ListingGenerator` object which yields :class:`.Comment` or
            :class:`.Submission` objects the user has upvoted.

        :raises: ``asyncprawcore.Forbidden`` if the user is not authorized to access the
            list.

            .. note::

                Since this function returns a :class:`.ListingGenerator` the exception
                may not occur until sometime after this function has returned.


        Additional keyword arguments are passed in the initialization of
        :class:`.ListingGenerator`.

        For example, to get all upvoted items of the authenticated user:

        .. code-block:: python

            current_user = await reddit.user.me()
            async for item in current_user.upvoted():
                print(item.id)

        """
        return ListingGenerator(
            self._reddit, urljoin(self._path, "upvoted"), **generator_kwargs
        )



import pytest

from asyncpraw.models.reddit.mixins import ThingModerationMixin

from .... import UnitTest


class TestThingModerationMixin(UnitTest):
    async def test_must_be_extended(self):
        with pytest.raises(NotImplementedError):
            await ThingModerationMixin().send_removal_message(
                message="public", title="title", type="message"
            )



"""Provide the ModNoteMixin class."""
from typing import TYPE_CHECKING, AsyncGenerator, Optional

if TYPE_CHECKING:  # pragma: no cover
    import asyncpraw


class ModNoteMixin:
    """Interface for classes that can have a moderator note set on them."""

    async def create_note(
        self, *, label: Optional[str] = None, note: str, **other_settings
    ) -> "asyncpraw.models.ModNote":
        """Create a moderator note on the author of this object in the subreddit it's posted in.

        :param label: The label for the note. As of this writing, this can be one of the
            following: ``"ABUSE_WARNING"``, ``"BAN"``, ``"BOT_BAN"``,
            ``"HELPFUL_USER"``, ``"PERMA_BAN"``, ``"SOLID_CONTRIBUTOR"``,
            ``"SPAM_WARNING"``, ``"SPAM_WATCH"``, or ``None`` (default: ``None``).
        :param note: The content of the note. As of this writing, this is limited to 250
            characters.
        :param other_settings: Additional keyword arguments are passed to
            :meth:`~.BaseModNotes.create`.

        :returns: The new :class:`.ModNote` object.

        For example, to create a note on a :class:`.Submission`, try:

        .. code-block:: python

            submission = await reddit.submission("92dd8")
            await submission.mod.create_note(label="HELPFUL_USER", note="Test note")

        """
        return await self.thing.subreddit.mod.notes.create(
            label=label, note=note, thing=self.thing, **other_settings
        )

    def author_notes(
        self, **generator_kwargs
    ) -> AsyncGenerator["asyncpraw.models.ModNote", None]:
        """Get the moderator notes for the author of this object in the subreddit it's posted in.

        :param generator_kwargs: Additional keyword arguments are passed in the
            initialization of the moderator note generator.

        :returns: A generator of :class:`.ModNote`.

        For example, to list all notes the author of a submission, try:

        .. code-block:: python

            for note in reddit.submission("92dd8").mod.author_notes():
                print(f"{note.label}: {note.note}")

        """
        return self.thing.subreddit.mod.notes.redditors(
            self.thing.author, **generator_kwargs
        )



"""Provide helper classes used by other models."""
import asyncio
import random
from collections import OrderedDict
from functools import wraps
from typing import Any, AsyncGenerator, Callable, List, Optional, Set, Union
from warnings import warn

from ..util import _deprecate_args


class BoundedSet:
    """A set with a maximum size that evicts the oldest items when necessary.

    This class does not implement the complete set interface.

    """

    def __init__(self, max_items: int):
        """Initialize a :class:`.BoundedSet` instance."""
        self.max_items = max_items
        self._set = OrderedDict()

    def __contains__(self, item: Any) -> bool:
        """Test if the :class:`.BoundedSet` contains item."""
        self._access(item)
        return item in self._set

    def _access(self, item: Any):
        if item in self._set:
            self._set.move_to_end(item)

    def add(self, item: Any):
        """Add an item to the set discarding the oldest item if necessary."""
        self._access(item)
        self._set[item] = None
        if len(self._set) > self.max_items:
            self._set.popitem(last=False)


class ExponentialCounter:
    """A class to provide an exponential counter with jitter."""

    def __init__(self, max_counter: int):
        """Initialize an :class:`.ExponentialCounter` instance.

        :param max_counter: The maximum base value.

            .. note::

                The computed value may be 3.125% higher due to jitter.

        """
        self._base = 1
        self._max = max_counter

    def counter(self) -> Union[int, float]:
        """Increment the counter and return the current value with jitter."""
        max_jitter = self._base / 16.0
        value = self._base + random.random() * max_jitter - max_jitter / 2
        self._base = min(self._base * 2, self._max)
        return value

    def reset(self):
        """Reset the counter to 1."""
        self._base = 1


def deprecate_lazy(func):  # noqa: D401
    """A decorator used for deprecating the ``lazy`` keyword argument."""

    @wraps(func)
    def wrapper(*args, **kwargs):
        if "lazy" in kwargs:
            kwargs.setdefault("fetch", not kwargs.pop("lazy"))
            warn(
                "The parameter ``lazy`` has been renamed to ``fetch`` and support for"
                " the ``lazy`` parameter will be removed in a future version of Async"
                " PRAW.",
                category=DeprecationWarning,
                stacklevel=3,
            )
        return func(*args, **kwargs)

    return wrapper


@_deprecate_args("permissions", "known_permissions")
def permissions_string(
    *, known_permissions: Set[str], permissions: Optional[List[str]]
) -> str:
    """Return a comma separated string of permission changes.

    :param known_permissions: A set of strings representing the available permissions.
    :param permissions: A list of strings, or ``None``. These strings can exclusively
        contain ``+`` or ``-`` prefixes, or contain no prefixes at all. When prefixed,
        the resulting string will simply be the joining of these inputs. When not
        prefixed, all permissions are considered to be additions, and all permissions in
        the ``known_permissions`` set that aren't provided are considered to be
        removals. When ``None``, the result is ``"+all"``.

    """
    if permissions is None:
        to_set = ["+all"]
    else:
        to_set = ["-all"]
        omitted = sorted(known_permissions - set(permissions))
        to_set.extend(f"-{x}" for x in omitted)
        to_set.extend(f"+{x}" for x in permissions)
    return ",".join(to_set)


@_deprecate_args(
    "function", "pause_after", "skip_existing", "attribute_name", "exclude_before"
)
async def stream_generator(
    function: Callable,
    *,
    attribute_name: str = "fullname",
    exclude_before: bool = False,
    pause_after: Optional[int] = None,
    skip_existing: bool = False,
    **function_kwargs: Any,
) -> AsyncGenerator[Any, None]:
    """Yield new items from ``function`` as they become available.

    :param function: A callable that returns a :class:`.ListingGenerator`, e.g.,
        :meth:`.Subreddit.comments` or :meth:`.Subreddit.new`.
    :param attribute_name: The field to use as an ID (default: ``"fullname"``).
    :param exclude_before: When ``True`` does not pass ``params`` to ``function``
        (default: ``False``).
    :param pause_after: An integer representing the number of requests that result in no
        new items before this function yields ``None``, effectively introducing a pause
        into the stream. A negative value yields ``None`` after items from a single
        response have been yielded, regardless of number of new items obtained in that
        response. A value of ``0`` yields ``None`` after every response resulting in no
        new items, and a value of ``None`` never introduces a pause (default: ``None``).
    :param skip_existing: When ``True``, this does not yield any results from the first
        request thereby skipping any items that existed in the stream prior to starting
        the stream (default: ``False``).

    Additional keyword arguments will be passed to ``function``.

    .. note::

        This function internally uses an exponential delay with jitter between
        subsequent responses that contain no new results, up to a maximum delay of just
        over 16 seconds. In practice, that means that the time before pause for
        ``pause_after=N+1`` is approximately twice the time before pause for
        ``pause_after=N``.

    For example, to create a stream of comment replies, try:

    .. code-block:: python

        reply_function = reddit.inbox.comment_replies
        async for reply in asyncpraw.models.util.stream_generator(reply_function):
            print(reply)

    To pause a comment stream after six responses with no new comments, try:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        async for comment in subreddit.stream.comments(pause_after=6):
            if comment is None:
                break
            print(comment)

    To resume fetching comments after a pause, try:

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        comment_stream = subreddit.stream.comments(pause_after=5)

        async for comment in comment_stream:
            if comment is None:
                break
            print(comment)
        # Do any other processing, then try to fetch more data
        async for comment in comment_stream:
            if comment is None:
                break
            print(comment)

    To bypass the internal exponential backoff, try the following. This approach is
    useful if you are monitoring a subreddit with infrequent activity, and you want to
    consistently learn about new items from the stream as soon as possible, rather than
    up to a delay of just over sixteen seconds.

    .. code-block:: python

        subreddit = await reddit.subreddit("test")
        async for comment in subreddit.stream.comments(pause_after=0):
            if comment is None:
                continue
            print(comment)

    """
    before_attribute = None
    exponential_counter = ExponentialCounter(max_counter=16)
    seen_attributes = BoundedSet(301)
    without_before_counter = 0
    responses_without_new = 0
    valid_pause_after = pause_after is not None
    while True:
        found = False
        newest_attribute = None
        limit = 100
        if before_attribute is None:
            limit -= without_before_counter
            without_before_counter = (without_before_counter + 1) % 30
        if not exclude_before:
            function_kwargs["params"] = {"before": before_attribute}
        for item in reversed(
            [result async for result in function(limit=limit, **function_kwargs)]
        ):
            attribute = getattr(item, attribute_name)
            if attribute in seen_attributes:
                continue
            found = True
            seen_attributes.add(attribute)
            newest_attribute = attribute
            if not skip_existing:
                yield item
        before_attribute = newest_attribute
        skip_existing = False
        if valid_pause_after and pause_after < 0:
            yield None
        elif found:
            exponential_counter.reset()
            responses_without_new = 0
        else:
            responses_without_new += 1
            if valid_pause_after and responses_without_new > pause_after:
                exponential_counter.reset()
                responses_without_new = 0
                yield None
            else:
                await asyncio.sleep(exponential_counter.counter())



import sys
from base64 import urlsafe_b64encode

import pytest
from asyncprawcore import Forbidden, NotFound

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import RedditAPIException
from asyncpraw.models import Redditor, WikiPage

from ... import IntegrationTest


class TestWikiPage(IntegrationTest):
    async def test_content_md(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        with self.use_cassette():
            page = await subreddit.wiki.get_page("test")
            assert page.content_md

    async def test_content_md__invalid_name(self):
        subreddit = await self.reddit.subreddit("reddit.com")
        page = WikiPage(self.reddit, subreddit, "\\A")

        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await page._fetch()
            assert str(excinfo.value) == "INVALID_PAGE_NAME"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_discussions(self, _):
        subreddit = await self.reddit.subreddit("reddit.com")

        with self.use_cassette():
            page = await subreddit.wiki.get_page("search")
            assert await self.async_list(page.discussions())

    async def test_edit(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        self.reddit.read_only = False
        with self.use_cassette():
            page = await subreddit.wiki.get_page("test")
            await page.edit(content="PRAW updated")

    async def test_edit__usernotes(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, "usernotes")
        with open("tests/integration/files/too_large.jpg", "rb") as fp:
            large_content = urlsafe_b64encode(fp.read()).decode()

        self.reddit.read_only = False
        with self.use_cassette():
            await page.edit(content=large_content)

    async def test_edit__with_reason(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        self.reddit.read_only = False
        with self.use_cassette():
            page = await subreddit.wiki.get_page("test")
            await page.edit(content="PRAW updated with reason", reason="PRAW testing")

    async def test_init__with_revision(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            page = WikiPage(
                self.reddit,
                subreddit,
                "index",
                revision="08a4ddfa-c000-11ea-999f-0e1fd5dedea1",
            )
            await page.load()
            assert isinstance(page.revision_by, Redditor)
            assert page.revision_date == 1594091696

    async def test_init__with_revision__author_deleted(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        page = WikiPage(
            self.reddit,
            subreddit,
            "config/stylesheet",
            revision="75390fec-8b8f-11e8-8a49-0edb077d29ac",
        )
        with self.use_cassette():
            await page.load()
            assert page.revision_by is None

    async def test_invalid_page(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        with self.use_cassette():
            with pytest.raises(NotFound):
                await subreddit.wiki.get_page("invalid")

    async def test_revision_by(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        with self.use_cassette():
            page = await subreddit.wiki.get_page("index")
            assert isinstance(page.revision_by, Redditor)

    async def test_revision(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        revision_id = "08a4ddfa-c000-11ea-999f-0e1fd5dedea1"

        with self.use_cassette():
            page = await subreddit.wiki.get_page("index")
            revision = await page.revision(revision_id)
            assert len(revision.content_md) > 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_revisions(self, _):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        with self.use_cassette():
            count = 0
            page = await subreddit.wiki.get_page("index")
            async for revision in page.revisions(limit=None):
                count += 1
                assert isinstance(revision["author"], Redditor)
                assert isinstance(revision["page"], WikiPage)
            assert count > 0

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_revisions__author_deleted(self, _):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        with self.use_cassette():
            page = await subreddit.wiki.get_page("config/submit_text")
            assert any(
                [
                    revision["author"] is None
                    async for revision in page.revisions(limit=10)
                ]
            )


class TestWikiPageModeration(IntegrationTest):
    async def test_add(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        self.reddit.read_only = False
        with self.use_cassette():
            page = await subreddit.wiki.get_page("index")
            await page.mod.add("Lil_SpazTest")

    async def test_remove(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        self.reddit.read_only = False
        with self.use_cassette():
            page = await subreddit.wiki.get_page("index")
            await page.mod.remove("Lil_SpazTest")

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_revert(self, _):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)
        page = WikiPage(self.reddit, subreddit, "test")

        self.reddit.read_only = False
        with self.use_cassette():
            revision_id = (await self.async_next(page.revisions(limit=1)))["id"]
            revision = await page.revision(revision_id)
            await revision.mod.revert()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_revert_css_fail(self, _):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        self.reddit.read_only = False
        with self.use_cassette():
            page = await subreddit.wiki.get_page("config/stylesheet")
            await subreddit.stylesheet.upload(
                name="css-revert-fail",
                image_path="tests/integration/files/icon.jpg",
            )
            await page.edit(content="div {background: url(%%css-revert-fail%%)}")
            revision_id = (await self.async_next(page.revisions(limit=1)))["id"]
            await subreddit.stylesheet.delete_image("css-revert-fail")
            with pytest.raises(Forbidden) as exc:
                revision = await page.revision(revision_id)
                await revision.mod.revert()
            assert await exc.value.response.json() == {
                "reason": "INVALID_CSS",
                "message": "Forbidden",
                "explanation": "%(css_error)s",
            }

    async def test_settings(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        self.reddit.read_only = False
        with self.use_cassette():
            page = await subreddit.wiki.get_page("index")
            settings = await page.mod.settings()
        assert {"editors": [], "listed": True, "permlevel": 0} == settings

    async def test_update(self):
        subreddit = await self.reddit.subreddit(pytest.placeholders.test_subreddit)

        self.reddit.read_only = False
        with self.use_cassette():
            page = await subreddit.wiki.get_page("index")
            updated = await page.mod.update(listed=False, permlevel=1)
        assert {"editors": [], "listed": False, "permlevel": 1} == updated



import sys
from datetime import datetime

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import ModmailMessage, Subreddit

from ... import IntegrationTest


class TestModmailConversation(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_archive(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.archive()
            conversation = await subreddit.modmail("faj6z")
            assert conversation.state == 2

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_highlight(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.highlight()
            conversation = await subreddit.modmail("faj6z")
            assert conversation.is_highlighted

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mute(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.mute()
            conversation = await subreddit.modmail("faj6z")
            assert conversation.user.mute_status["isMuted"]

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_mute_duration(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("g46rw")
            await conversation.mute(num_days=7)
            conversation = await subreddit.modmail("g46rw")
            assert conversation.user.mute_status["isMuted"]
            diff = datetime.fromisoformat(
                conversation.user.mute_status["endDate"]
            ) - datetime.fromisoformat(conversation.mod_actions[-1].date)
            assert diff.days == 6  # 6 here because it is not 7 whole days

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_read(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.read()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_read__other_conversations(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            other_conversation = await subreddit.modmail("fajcu")
            await conversation.read(other_conversations=[other_conversation])

    async def test_reply(self):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            reply = await conversation.reply(body="A message")
        assert isinstance(reply, ModmailMessage)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unarchive(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.unarchive()
            conversation = await subreddit.modmail("faj6z")
            assert conversation.state == 1

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unhighlight(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.unhighlight()
            conversation = await subreddit.modmail("faj6z")
            assert not conversation.is_highlighted

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unmute(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.unmute()
            conversation = await subreddit.modmail("faj6z")
            assert not conversation.user.mute_status["isMuted"]

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_unread(self, _):
        self.reddit.read_only = False
        subreddit = Subreddit(self.reddit, "all")
        with self.use_cassette():
            conversation = await subreddit.modmail("faj6z")
            await conversation.unread()
            conversation = await subreddit.modmail("faj6z")
            assert conversation.last_unread is not None



"""Test asyncpraw.models.auth."""
import pytest
from asyncprawcore import InvalidToken

from asyncpraw import Reddit

from .. import IntegrationTest


class TestAuthScript(IntegrationTest):
    async def test_scopes(self):
        with self.use_cassette():
            assert self.reddit.read_only is True
            assert await self.reddit.auth.scopes() == {"*"}


class TestAuthWeb(IntegrationTest):
    def setup_reddit(self):
        self.reddit = Reddit(
            client_id=pytest.placeholders.client_id,
            client_secret=pytest.placeholders.client_secret,
            redirect_uri=pytest.placeholders.redirect_uri,
            user_agent=pytest.placeholders.user_agent,
            username=None,
        )

    async def test_authorize(self):
        with self.use_cassette():
            token = await self.reddit.auth.authorize(pytest.placeholders.auth_code)
            assert isinstance(token, str)
            assert self.reddit.read_only is False
            assert await self.reddit.auth.scopes() == {"submit"}

    async def test_scopes__read_only(self):
        with self.use_cassette():
            assert self.reddit.read_only is True
            assert await self.reddit.auth.scopes() == {"*"}


class TestAuthImplicit(IntegrationTest):
    def setup_reddit(self):
        self.reddit = Reddit(
            client_id=pytest.placeholders.client_id,
            client_secret=None,
            user_agent=pytest.placeholders.user_agent,
        )

    async def test_implicit__with_invalid_token(self):
        self.reddit.auth.implicit(
            access_token="invalid_token", expires_in=10, scope="read"
        )
        with self.use_cassette():
            with pytest.raises(InvalidToken):
                await self.reddit.user.me()

    async def test_scopes__read_only(self):
        with self.use_cassette():
            assert self.reddit.read_only is True
            assert await self.reddit.auth.scopes() == {"*"}



"""Test asyncpraw.models.Trophy."""
import pytest

from asyncpraw.models.trophy import Trophy

from .. import UnitTest


class TestTrophy(UnitTest):
    def test_init__reject_no_name(self, reddit):
        with pytest.raises(AssertionError):
            Trophy(reddit, {})
        with pytest.raises(AssertionError):
            Trophy(reddit, {"id": "abcd"})

    def test_init__reject_non_dict(self, reddit):
        with pytest.raises(AssertionError):
            Trophy(reddit, None)
        with pytest.raises(AssertionError):
            Trophy(reddit, "")

    def test_init__str_returns_name(self, reddit):
        name = "Inciteful Link"
        trophy = Trophy(reddit, {"name": name})
        assert str(trophy) == trophy.name == name

    def test_equality(self, reddit):
        name = "Inciteful Link"
        trophy = Trophy(reddit, {"name": name})
        assert trophy != 5

    def test_repr(self, reddit):
        name = "Inciteful Link"
        trophy = Trophy(reddit, {"name": name})
        assert repr(trophy) == "Trophy(name='Inciteful Link')"



import sys

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import Comment, Submission, Subreddit

from ... import IntegrationTest


class TestMultireddit(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_add(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            multireddits = await self.reddit.user.multireddits()
            multi = multireddits[0]
            await multi.add("redditdev")
            await multi.load()
            assert "redditdev" in multi.subreddits

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_copy(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            multi = await self.reddit.multireddit(
                redditor="kjoneslol", name="sfwpornnetwork", fetch=True
            )
            new = await multi.copy()
        assert new.name == multi.name
        assert new.display_name == multi.display_name
        assert "kjoneslol" not in new.path

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_copy__with_display_name(self, _):
        self.reddit.read_only = False
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        name = "A--B\n" * 10
        with self.use_cassette():
            new = await multi.copy(display_name=name)
        assert new.name == "a_b_a_b_a_b_a_b_a_b"
        assert new.display_name == name
        assert "kjoneslol" not in new.path

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_create(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            multireddit = await self.reddit.multireddit.create(
                display_name="Async PRAW create test", subreddits=["redditdev"]
            )
        assert multireddit.display_name == "Async PRAW create test"
        assert multireddit.name == "async_praw_create_test"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_delete(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            multireddits = await self.reddit.user.multireddits()
            multi = multireddits[0]
            await multi.delete()

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_remove(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            multireddits = await self.reddit.user.multireddits()
            multi = multireddits[0]
            await multi.remove("redditdev")
            await multi.load()
            assert "redditdev" not in multi.subreddits

    async def test_subreddits(self):
        with self.use_cassette():
            multi = await self.reddit.multireddit(
                redditor="kjoneslol", name="sfwpornnetwork", fetch=True
            )
            assert multi.subreddits
        assert all(isinstance(x, Subreddit) for x in multi.subreddits)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_update(self, _):
        self.reddit.read_only = False
        subreddits = ["pokemongo", "pokemongodev"]
        with self.use_cassette():
            multireddits = await self.reddit.user.multireddits()
            multi = multireddits[0]
            prev_path = multi.path
            await multi.update(
                display_name="Updated display name", subreddits=subreddits
            )
        assert multi.display_name == "Updated display name"
        assert multi.path == prev_path
        assert multi.subreddits == subreddits


class TestMultiredditListings(IntegrationTest):
    async def test_comments(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            comments = await self.async_list(multi.comments())
        assert len(comments) == 100

    async def test_controversial(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            submissions = await self.async_list(multi.controversial())
        assert len(submissions) == 100

    async def test_gilded(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            submissions = await self.async_list(multi.gilded())
        assert len(submissions) == 100

    async def test_hot(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            submissions = await self.async_list(multi.hot())
        assert len(submissions) == 100

    async def test_new(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            submissions = await self.async_list(multi.new())
        assert len(submissions) == 100

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_new__self_multi(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            multireddits = await self.reddit.user.multireddits()
            multi = multireddits[0]
            submissions = await self.async_list(multi.new())
        assert len(submissions) == 100

    async def test_random_rising(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            submissions = await self.async_list(multi.random_rising())
        assert len(submissions) > 0

    async def test_rising(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            submissions = await self.async_list(multi.rising())
        assert len(submissions) > 0

    async def test_top(self):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            submissions = await self.async_list(multi.top())
        assert len(submissions) == 100


class TestMultiredditStreams(IntegrationTest):
    @mock.patch("asyncio.sleep", return_value=None)
    async def test_comments(self, _):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            generator = multi.stream.comments()
            for i in range(101):
                assert isinstance(await self.async_next(generator), Comment)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_comments__with_pause(self, _):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            comment_stream = multi.stream.comments(pause_after=0)
            comment_count = 1
            pause_count = 1
            comment = await self.async_next(comment_stream)
            while comment is not None:
                comment_count += 1
                comment = await self.async_next(comment_stream)
            while comment is None:
                pause_count += 1
                comment = await self.async_next(comment_stream)
            assert comment_count == 101
            assert pause_count == 2

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submissions(self, _):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette():
            generator = multi.stream.submissions()
            for i in range(102):
                assert isinstance(await self.async_next(generator), Submission)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_submissions__with_pause(self, _):
        multi = await self.reddit.multireddit(
            redditor="kjoneslol", name="sfwpornnetwork"
        )
        with self.use_cassette("TestMultiredditStreams.test_submissions"):
            generator = multi.stream.submissions(pause_after=-1)
            submission = await self.async_next(generator)
            submission_count = 0
            while submission is not None:
                submission_count += 1
                submission = await self.async_next(generator)
            assert submission_count == 100



"""Test asyncpraw.models.redditors."""
import sys

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.models import Redditor, Subreddit

from .. import IntegrationTest


class TestRedditors(IntegrationTest):
    async def test_new(self):
        with self.use_cassette():
            profiles = await self.async_list(self.reddit.redditors.new(limit=300))
        assert len(profiles) == 300
        assert all(isinstance(profile, Subreddit) for profile in profiles)
        assert all(str(profile).startswith("u_") for profile in profiles)

    async def test_popular(self):
        with self.use_cassette():
            profiles = await self.async_list(self.reddit.redditors.popular(limit=15))
        assert len(profiles) == 15
        assert all(isinstance(profile, Subreddit) for profile in profiles)
        assert all(str(profile).startswith("u_") for profile in profiles)

    async def test_search(self):
        with self.use_cassette():
            found = False
            async for profile in self.reddit.redditors.search("praw"):
                assert isinstance(profile, Redditor)
                found = True
            assert found

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_stream(self, _):
        with self.use_cassette():
            generator = self.reddit.redditors.stream()
            for i in range(101):
                assert isinstance(await self.async_next(generator), Subreddit)

    async def test_partial_redditors(self):
        with self.use_cassette():
            gen = self.reddit.redditors.partial_redditors(["t2_1w72", "t2_4x25quk"])
            user_data = await self.async_list(gen)

        fullnames = [user.fullname for user in user_data]
        assert fullnames == ["t2_1w72", "t2_4x25quk"]
        assert user_data[0].fullname == "t2_1w72"
        assert user_data[0].name == "spez"

    async def test_partial_redditors__not_found(self):
        with self.use_cassette():
            gen = self.reddit.redditors.partial_redditors(
                ["t2_invalid_abc", "t2_invalid_123"]
            )
            assert await self.async_list(gen) == []

            gen = self.reddit.redditors.partial_redditors(
                ["t2_invalid_abc" for _ in range(100)] + ["t2_4x25quk"]
            )
            assert [user.fullname async for user in gen] == ["t2_4x25quk"]



import pytest

from asyncpraw.models import Message, SubredditMessage

from ... import UnitTest


class TestMessage(UnitTest):
    def test_attribute_error(self, reddit):
        with pytest.raises(AttributeError):
            Message(reddit, _data={"id": "1"}).mark_as_read()

    def test_equality(self, reddit):
        message1 = Message(reddit, _data={"id": "1"})
        message2 = Message(reddit, _data={"id": "1"})
        message3 = Message(reddit, _data={"id": "2"})
        assert message1 == message1
        assert message2 == message2
        assert message3 == message3
        assert message1 == message2
        assert message2 != message3
        assert message1 != message3
        assert "1" == message1
        assert message1 == "1"

    def test_fullname(self, reddit):
        message = Message(reddit, _data={"id": "dummy"})
        assert message.fullname == "t4_dummy"

    def test_hash(self, reddit):
        message1 = Message(reddit, _data={"id": "dummy1"})
        message2 = Message(reddit, _data={"id": "dummy1"})
        message3 = Message(reddit, _data={"id": "dummy2"})
        assert hash(message1) == hash(message1)
        assert hash(message2) == hash(message2)
        assert hash(message3) == hash(message3)
        assert hash(message1) == hash(message2)
        assert hash(message2) != hash(message3)
        assert hash(message1) != hash(message3)

    def test_repr(self, reddit):
        message = Message(reddit, _data={"id": "dummy"})
        assert repr(message) == "Message(id='dummy')"

    def test_str(self, reddit):
        message = Message(reddit, _data={"id": "dummy"})
        assert str(message) == "dummy"


class TestSubredditMessage(UnitTest):
    def test_inheritance(self, reddit):
        message = SubredditMessage(reddit, _data={"id": "dummy"})
        assert isinstance(message, Message)

    def test_repr(self, reddit):
        message = SubredditMessage(reddit, _data={"id": "dummy"})
        assert repr(message) == "SubredditMessage(id='dummy')"

    def test_str(self, reddit):
        message = SubredditMessage(reddit, _data={"id": "dummy"})
        assert str(message) == "dummy"



"""Test asyncpraw.reddit."""
import sys
from base64 import urlsafe_b64encode

import pytest
from asyncprawcore.exceptions import BadRequest, ServerError

if sys.version_info < (3, 8):
    from asynctest import mock
else:
    from unittest import mock

from asyncpraw.exceptions import RedditAPIException
from asyncpraw.models import LiveThread
from asyncpraw.models.reddit.base import RedditBase
from asyncpraw.models.reddit.submission import Submission
from asyncpraw.models.reddit.subreddit import Subreddit

from . import IntegrationTest


class TestReddit(IntegrationTest):
    async def test_bad_request_without_json_text_plain_response(self):
        with open("tests/integration/files/too_large.jpg", "rb") as fp:
            junk = urlsafe_b64encode(fp.read()).decode()
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await self.reddit.request(
                    method="GET",
                    path=f"/api/morechildren?link_id=t3_n7r3uz&children={junk}",
                )
            assert str(excinfo.value) == "Bad Request"

    async def test_bad_request_without_json_text_html_response(self):
        with open("tests/integration/files/comment_ids.txt") as fp:
            ids = fp.read()[:8000]
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as excinfo:
                await self.reddit.request(
                    method="GET",
                    path=f"/api/morechildren?link_id=t3_n7r3uz&children={ids}",
                )
            assert (
                str(excinfo.value)
                == "<html><body><h1>400 Bad request</h1>\nYour browser sent an invalid"
                " request.\n</body></html>\n"
            )

    async def test_bare_badrequest(self):
        data = {
            "sr": "AskReddit",
            "field": "link",
            "kind": "link",
            "title": "l",
            "text": "lol",
            "show_error_list": True,
        }
        self.reddit.read_only = False
        with self.use_cassette():
            with pytest.raises(BadRequest):
                await self.reddit.post("/api/validate_submission_field", data=data)

    async def test_info(self):
        bases = ["t1_d7ltv", "t3_5dec", "t5_2qk"]
        items = []
        for i in range(100):
            for base in bases:
                items.append(f"{base}{i:02d}")

        with self.use_cassette():
            results = await self.async_list(self.reddit.info(fullnames=items))
        assert len(results) > 100
        for item in results:
            assert isinstance(item, RedditBase)

    async def test_info_url(self):
        with self.use_cassette():
            results = await self.async_list(self.reddit.info(url="youtube.com"))
        assert len(results) > 0
        for item in results:
            assert isinstance(item, Submission)

    async def test_info_sr_names(self):
        items = [await self.reddit.subreddit("redditdev"), "reddit.com", "t:1337", "nl"]
        item_generator = self.reddit.info(subreddits=items)
        with self.use_cassette():
            results = await self.async_list(item_generator)
        assert len(results) == 4
        for item in results:
            assert isinstance(item, Subreddit)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_live_call(self, _):
        thread_id = "ukaeu1ik4sw5"
        with self.use_cassette():
            thread = await self.reddit.live(thread_id, fetch=True)
            assert thread.title == "reddit updates"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_live_create(self, _):
        self.reddit.read_only = False
        with self.use_cassette():
            live = await self.reddit.live.create("PRAW Create Test")
            assert isinstance(live, LiveThread)
            await live._fetch()
            assert live.title == "PRAW Create Test"

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_live_info__contain_invalid_id(self, _):
        gen = self.reddit.live.info(["invalid"])
        with self.use_cassette():
            with pytest.raises(ServerError):
                await self.async_list(gen)

    @mock.patch("asyncio.sleep", return_value=None)
    async def test_live_info(self, _):
        ids = """
        ta40aifzobnv ta40l9u2ermf ta40ucdiq366 ta416hjgvbhy ta41ln5vsyaz
        ta42cyzy1nze ta42i49806k0 ta436ojd653m ta43945wgmaa ta43znjvza9t
        ta4418kxie3z ta44mk0nllhm ta45j0yvww9t ta4613lzdh8q ta46l8k86jt9
        ta47qua0xu3n ta489fm9515p ta48ml5k1uk9 ta48zy4jzjcb ta49irvwndau
        ta49upckgoyw ta4a02h9ynsb ta4aa4lrgvst ta4alauoi8ws ta4aqyacr70u

        ta4ekdk6m5g2 ta4ezvoc49gy ta4f3iv06c1n ta4ffvliq5l7 ta4fib9lx3zx
        ta4gka0ll41h ta4h89f6isfg ta4ht7s8he49 ta4i1eb564ar ta4imxhap4fg
        ta4iu3g9whtk ta4j3o05j0d3 ta4kloqi6csg ta4m6kj44dql ta4mlqtihiil
        ta4ng30l3fz1 ta4nldsjimhu ta4pd78tuk29 ta4prwyy1w9i ta4pvu8y6f8o
        ta4ray2odqub ta4rua4oe6a1 ta4tk9fwjgz1 ta4trgqw6mmx ta4tv3sen7u4

        ta4uyh0fnc0a ta4v54gnggcl ta4v5cm004z1 ta4vortaefna ta4wqym9d0v3
        ta4wsuouxjtm ta4x7jr9v0fn ta4yast5e96b ta4z337yzlgu ta4zo9zzo9ui
        ta507u2euo3w ta50exn0mtx1 ta51x2crezff ta52ch48gn6l ta53ijowvc6z
        ta53iy196uod ta541cz1hfb4 ta54n0ncx8pc ta55ytfmre2g ta581bybyjwi
        ta59gcidn2ym ta59mkwnrd43 ta5ar22wzi2w ta5awwo42ibb ta5dhmvylw0l

        ta5hipd6wahr ta5itb7clg3s ta5nlm09y8kb ta5nm0f831x1 ta5oavbflorf
        ta5rnv18s85o ta5ru6ysh254 ta5sfz02nc8b ta5syawj086b ta5t41osygln
        ta5uy3ynoo4a ta5w0seb1xfy ta5wddbh0ln0 ta5zmjzuijwo ta617ozbmxhb
        ta64q6pjz2bs ta696fdie4ne ta6bmog7gvoq ta6f9y7sdzru ta6j838d2wjn
        ta6l4q5c17fd ta6ofypk3yp2 ta6sjmjt1aeb ta6sqhgyv41q ta70eezhz50r

        ta72azs1l4u9 ta74r3dp2pt5 ta7pfcqdx9cl ta8zxbt2sk6z ta94nde51q4i
        """.split()
        gen = self.reddit.live.info(ids)
        with self.use_cassette():
            threads = await self.async_list(gen)
        assert len(threads) > 100
        assert all(isinstance(thread, LiveThread) for thread in threads)
        thread_ids = [thread.id for thread in threads]
        assert thread_ids == ids

    # async def test_live_now__featured(self): # TODO: record when something is featured
    #     with self.use_cassette():
    #         thread = await self.reddit.live.now()
    #     assert isinstance(thread, LiveThread)
    #     assert thread.id == "z2f981agq7ky"

    async def test_live_now__no_featured(self):
        with self.use_cassette():
            now = await self.reddit.live.now()
            assert now is None

    async def test_notes__call__(self):
        self.reddit.read_only = False
        with self.use_cassette():
            notes = await self.async_list(
                self.reddit.notes(
                    pairs=[
                        (
                            await self.reddit.subreddit(
                                pytest.placeholders.test_subreddit
                            ),
                            "Watchful1",
                        ),
                        (
                            pytest.placeholders.test_subreddit,
                            await self.reddit.redditor("Watchful12"),
                        ),
                        (pytest.placeholders.test_subreddit, "spez"),
                    ],
                    things=[await self.reddit.submission("uflrmv")],
                )
            )
            assert len(notes) == 4
            assert notes[0].user.name.lower() == "watchful1"
            assert notes[1].user.name.lower() == "watchful12"
            assert notes[2] is None

    async def test_notes__things(self):
        self.reddit.read_only = False
        with self.use_cassette():
            thing = await self.reddit.submission("uflrmv")
            notes = await self.async_list(self.reddit.notes.things(thing))
            assert len(notes) == 9
            assert notes[0].user == thing.author

    async def test_random_subreddit(self):
        names = set()
        with self.use_cassette():
            for i in range(3):
                sub = await self.reddit.random_subreddit()
                names.add(sub.display_name)
        assert len(names) == 3

    async def test_subreddit_with_randnsfw(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("randnsfw")
            assert subreddit.display_name != "randnsfw"
            assert subreddit.over18

    async def test_subreddit_with_random(self):
        with self.use_cassette():
            subreddit = await self.reddit.subreddit("random")
            assert subreddit.display_name != "random"

    async def test_username_available__available(self):
        fake_user = "prawtestuserabcd1234"
        with self.use_cassette():
            assert await self.reddit.username_available(fake_user)

    async def test_username_available__unavailable(self):
        with self.use_cassette():
            assert not await self.reddit.username_available("bboe")

    async def test_username_available_exception(self):
        with self.use_cassette():
            with pytest.raises(RedditAPIException) as exc:
                await self.reddit.username_available("a")
            assert str(exc.value) == "BAD_USERNAME: 'invalid user name' on field 'user'"


class TestDomainListing(IntegrationTest):
    async def test_controversial(self):
        with self.use_cassette():
            submissions = await self.async_list(
                self.reddit.domain("youtube.com").controversial()
            )
        assert len(submissions) == 100

    async def test_hot(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.domain("youtube.com").hot())
        assert len(submissions) == 100

    async def test_new(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.domain("youtube.com").new())
        assert len(submissions) == 100

    async def test_random_rising(self):
        with self.use_cassette():
            submissions = await self.async_list(
                self.reddit.domain("youtube.com").random_rising()
            )
        assert len(submissions) == 100

    async def test_rising(self):
        with self.use_cassette():
            await self.async_list(self.reddit.domain("youtube.com").rising())

    async def test_top(self):
        with self.use_cassette():
            submissions = await self.async_list(self.reddit.domain("youtube.com").top())
        assert len(submissions) == 100
